package com.ibm.btools.blm.ie.exprt;
/*
* Licensed Materials - Property of IBM
* 5724-I74, 5724-I75
*  Copyright IBM Corporation 2004, 2009. All rights reserved.
* US Government Users Restricted Rights - Use, duplication or disclosure
* restricted by GSA ADP Schedule Contract with IBM Corp.
*/
import java.lang.reflect.InvocationTargetException;
import java.util.List;
import java.util.Map;

import org.eclipse.core.runtime.IProgressMonitor;
import org.eclipse.core.runtime.OperationCanceledException;
import org.eclipse.ui.actions.WorkspaceModifyOperation;

import com.ibm.btools.blm.ie.imprt.engine.ProgressCalculator;
import com.ibm.btools.blm.ie.imprt.rule.util.IImportResultMessages;
import com.ibm.btools.blm.ie.imprt.rule.util.LoggingUtil;
import com.ibm.btools.blm.ie.imprt.rule.util.NavigatorUtil;
import com.ibm.btools.blm.ie.resource.IeMessageKeys;
import com.ibm.btools.util.exception.BTException;

/**
 * Export Engine which calls the export operation to export objects from the workspace to the local
 * file system.
 */
public class ExportEngine extends WorkspaceModifyOperation{
	// copyright	
	static final String copyright = "Licensed Material - Property of IBM " + " 5724-I74, 5724-I75 " + "(C) Copyright IBM Corporation 2004, 2009. All Rights Reserved. " + "U.S. Government Users Restricted Rights - Use, duplication or disclosure " + "restricted by GSA ADP Schedule Contract with IBM Corp.".intern(); //$NON-NLS-1$ //$NON-NLS-2$ //$NON-NLS-3$ //$NON-NLS-4$ //$NON-NLS-5$

	private String operationId;
	private String projectName;
	private IExportQuery queryCallback;
	private ExportResult result;
	private List modelElements;
	private String outputDir;	
	private boolean alwaysOverwrite = false;
	private Map additionalOptions;
	
	/**
	 * Constructor of ExportEngine
	 * @param operationId The id that represents the export Operation and described in 
	 *        the plugin extension 
	 * @param projectName The name of the destination.
	 * @param queryImplementor ExportQuery that will be implemented by the 
	 *        Wizard, so that we can get the feedback from the user.
	 * @param selectedModelElements The list of selected BOM model elements for export.
	 * @param outputDir output directory for the export
	 * @param options any options specific to the type of export
	 */
	public ExportEngine(final String operationId, final String projectName, 
		final IExportQuery queryImplementor, final List selectedModelElements, final String outputDir, 
		final Map options) {
		
		//super();
		this.projectName = projectName;
		this.operationId = operationId;
		queryCallback = queryImplementor;
		this.modelElements = selectedModelElements;		
		this.additionalOptions = options;
		this.outputDir = outputDir;		
	}
	
	/**
	 * Invokes the export operation 
	 */
	protected void execute(IProgressMonitor monitor) throws InvocationTargetException {
		LoggingUtil.traceEntry(this, "execute");		
		//defect 84298	
	//	monitor = progressMonitor;		
		IExportOperation exportOperation = null;	
		
		if (monitor != null) {			
			monitor.beginTask(null, ProgressCalculator.getTotalProgress());
		}		
		
		try {
//			if (projectName == null) {
//				LoggingUtil.logError(IeMessageKeys.GENERIC_ERROR, IImportResultMessages.PROJECT_NAME_IS_NULL);
//				LoggingUtil.traceExit(this, "execute");
//				return;
//			}
			if (operationId == null) {				
				LoggingUtil.logError(IeMessageKeys.GENERIC_ERROR, IImportResultMessages.OPERATION_ID_IS_NULL);
				LoggingUtil.traceExit(this, "execute");				
				return;
			}
			
			// if the queryCallback is null, we will handle any naming collision as ignore.			
			
			if (modelElements != null) {
				LoggingUtil.trace(this, "execute", "Selected nodes is not null.");
								
				ExportOperationRegistryReader reader = ExportOperationRegistryReader.getReader();
				exportOperation = reader.findExportOperation(operationId);
				
				if (exportOperation == null) {
					BTException e = new BTException(IeMessageKeys.GENERIC_ERROR);
					LoggingUtil.logError(IeMessageKeys.GENERIC_ERROR, IImportResultMessages.OPERATION_IS_NULL);					
					throw new InvocationTargetException(e);
				}
				LoggingUtil.trace(this, "execute", "Export Operation is loaded.");				
				
				// Create the export session				
				ExportSession session = new ExportSession();				
				session.setProgressMonitor(monitor);
				session.setExportQuery(queryCallback);
				//AZ add the project node for the session
				session.setProjectNode(NavigatorUtil.getProjectNode(projectName));
				
				ExportOptions options = new ExportOptions();
				options.setAlwaysOverwrite(alwaysOverwrite);
				if (additionalOptions != null) {
					LoggingUtil.trace(this, "execute", "Export Operation has additional options.");			
					options.setAdditionalOptions(additionalOptions);
				}
				session.setExportOption(options);								
				exportOperation.setModelElements(modelElements);
				exportOperation.setProjectName(projectName);
				exportOperation.setExportSession(session);
				exportOperation.setOutputDir(outputDir);
			
				
				// Begin
				if (monitor != null) {			
					monitor.worked(ProgressCalculator.getSimpleTask());
				}
				
				exportOperation.export();
				LoggingUtil.trace(this, "execute", "Export Operation is done.");	
													
				//set the result so UI can get it	
				result = exportOperation.getExportSession().getExportResult();
			}
		} catch (OperationCanceledException e) {
			LoggingUtil.logWarning(IeMessageKeys.CANCEL_THE_OPERATION, null, e, IImportResultMessages.RUNTIME_EXCEPTION);		

			throw e; 
		} catch (RuntimeException e) {
			LoggingUtil.logError(IeMessageKeys.GENERIC_ERROR, null, e, IImportResultMessages.RUNTIME_EXCEPTION);
		
			throw new InvocationTargetException(e);
		} finally {
			// The result of the engine need to be returned.			
			if (monitor != null) {	
				monitor.done();
			}			
		}
		LoggingUtil.traceExit(this, "execute");
	}

	/**
	 * Gets Export result
	 * @return ExportResult-the result of the export
	 */
	public ExportResult getExportResult() {
		return result;
	}

	/**
	 * Sets always overwrite
	 * @param alwaysOverwrite
	 */
	public void setAlwaysOverwrite(final boolean alwaysOverwrite) {
		this.alwaysOverwrite = alwaysOverwrite;
	}

	/**
	 * Gets output directory
	 * @return
	 */
	public String getOutputDir() {
		return outputDir;
	}

	/**
	 * sets the output directory
	 * @param string
	 */
	public void setOutputDir(String string) {
		outputDir = string;
	}

}