package com.ibm.btools.blm.ie.imprt.engine;
/*
* Licensed Materials - Property of IBM
* 5724-I74, 5724-I75
*  Copyright IBM Corporation 2004, 2009. All rights reserved.
* US Government Users Restricted Rights - Use, duplication or disclosure
* restricted by GSA ADP Schedule Contract with IBM Corp.
*/

import java.lang.reflect.InvocationTargetException;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;
import java.util.Map;

import org.eclipse.core.runtime.IProgressMonitor;
import org.eclipse.core.runtime.OperationCanceledException;
import org.eclipse.emf.common.util.EList;
import org.eclipse.ui.actions.WorkspaceModifyOperation;

import com.ibm.btools.blm.ie.imprt.IImportFramework;
import com.ibm.btools.blm.ie.imprt.IImportOperation;
import com.ibm.btools.blm.ie.imprt.IImportQuery;
import com.ibm.btools.blm.ie.imprt.ImportMessage;
import com.ibm.btools.blm.ie.imprt.ImportOptions;
import com.ibm.btools.blm.ie.imprt.ImportResult;
import com.ibm.btools.blm.ie.imprt.ImportSession;
import com.ibm.btools.blm.ie.imprt.registry.ImportOperationRegistryReader;
import com.ibm.btools.blm.ie.imprt.rule.util.BOMUtil;
import com.ibm.btools.blm.ie.imprt.rule.util.IImportResultMessages;
import com.ibm.btools.blm.ie.imprt.rule.util.LoggingUtil;
import com.ibm.btools.blm.ie.imprt.rule.util.NavigatorUtil;
import com.ibm.btools.blm.ie.resource.IeMessageKeys;
import com.ibm.btools.blm.ui.importExport.BLMImportExportConstants;
import com.ibm.btools.model.filemanager.FileMGR;
import com.ibm.btools.model.modelmanager.ModelMGR;
import com.ibm.btools.model.modelmanager.RemoveProjectCmd;
import com.ibm.btools.model.modelmanager.dependencymanager.DependencyManager;
import com.ibm.btools.model.modelmanager.dependencymanager.SaveDependencyModelCmd;
import com.ibm.btools.model.modelmanager.validation.BTReporter;
import com.ibm.btools.model.modelmanager.validation.BTValidator;
import com.ibm.btools.model.modelmanager.validation.ValidationException;
import com.ibm.btools.util.exception.BTException;


/**
 * Import Engine which call the import operation to import objects from the local
 * file system into the workspace.
 */
public class ImportEngine extends WorkspaceModifyOperation {
	// copyright	
	static final String copyright = "Licensed Material - Property of IBM " + " 5724-I74, 5724-I75 " + "(C) Copyright IBM Corporation 2004, 2009. All Rights Reserved. " + "U.S. Government Users Restricted Rights - Use, duplication or disclosure " + "restricted by GSA ADP Schedule Contract with IBM Corp.".intern(); //$NON-NLS-1$ //$NON-NLS-2$ //$NON-NLS-3$ //$NON-NLS-4$ //$NON-NLS-5$

	private IProgressMonitor monitor;

	private String operationId;
	private String projectName;
	private IImportQuery queryCallback;
	private ImportResult result;
	private List selectedFiles;	
	private boolean alwaysOverwrite = false;
	private boolean alwaysNewCopy = false;
	private Map additionalOptions;
	private boolean isEnabled = true;
	private boolean isProjCleanUpReq = false;
	
	/**
	 * Constructor of ImportEngine
	 * @param operationId The id that represents the Import Operation and described in 
	 *        the plugin extension 
	 * @param projectName The name of the destination.
	 * @param queryImplementor ImportQuery that will be implemented by the 
	 *        Wizard, so that we can get the feedback from the user.
	 * @param selectedFiles The list of selected files for import.
	 * @param options The list of options pass to the import operation
	 * @param isProjCleanUpReq A flag to indicate a project clean up from memory is necessary
	 */
	public ImportEngine(final String operationId, final String projectName, 
		final IImportQuery queryImplementor, final List selectedFiles, 
		final Map options, final boolean isProjCleanUpReq) {
		// The Eclipse ImportOperation uses an Import Structure Provider 
		// instead of hanlding the list of the files. The assumption here is
		// we only handle files from file system and we don't need to consider
		// the folder structure. Therefore, we don't need to follow the same
		// pattern.
		super();
		this.projectName = projectName;
		this.operationId = operationId;
		queryCallback = queryImplementor;
		this.selectedFiles = selectedFiles;		
		this.additionalOptions = options;
		this.isProjCleanUpReq = isProjCleanUpReq;
	}
	
	/**
	 * Validate the operation
	 */
	public void validate() {
		IImportOperation importOperation = null;
		result = null;
		try {		
			if (projectName == null) {
				LoggingUtil.logError(IeMessageKeys.GENERIC_ERROR, IImportResultMessages.PROJECT_NAME_IS_NULL);
				LoggingUtil.traceExit(this, "execute");
				return;
			}
			if (operationId == null) {				
				LoggingUtil.logError(IeMessageKeys.GENERIC_ERROR, IImportResultMessages.OPERATION_ID_IS_NULL);
				LoggingUtil.traceExit(this, "execute");				
				return;
			}
			ImportOperationRegistryReader reader = ImportOperationRegistryReader.getReader();
			importOperation = reader.findImportOperation(operationId);
				
			if (importOperation == null) {
				LoggingUtil.logError(IeMessageKeys.GENERIC_ERROR, IImportResultMessages.OPERATION_IS_NULL);
				LoggingUtil.traceExit(this, "execute");
				return;
			}			

			// Create the import session				
			ImportSession session = new ImportSession();				
			session.setImportQuery(queryCallback);

			ImportOptions options = new ImportOptions();
			options.setAlwaysOverwrite(alwaysOverwrite);
			options.setAlwaysNewCopy(alwaysNewCopy);
			if (additionalOptions != null) {
				LoggingUtil.trace(this, "execute", "Import Operation has additional options.");			
				options.setAdditionalOptions(additionalOptions);
			}
			session.setImportOption(options);								
			importOperation.setInputFiles(selectedFiles);
			importOperation.setProjectName(projectName);
			importOperation.setImportSession(session);

			importOperation.canExecute();		
			
		} finally {
			result = importOperation.getImportSession().getImportResult();
			if (result == null) {
				result = new ImportResult();
				result.addMessage(LoggingUtil.getMessage(IeMessageKeys.GENERIC_ERROR));			
			}
		}
	}	
	
	/**
	 * Invoke the Import operation 
	 */
	protected void execute(IProgressMonitor progressMonitor) throws InvocationTargetException {
		LoggingUtil.traceEntry(this, "execute");		
    	boolean isInfraEnabled = false; 
		result = null;
		
		BOMProcessor processor = null;
		BOMUpdateProcessor updateProcessor = null;
		BOMSimulationProcessor simProcessor=null;
		ImportSession session  = null;
		
		BTValidator validator = BTValidator.instance();
		
		monitor = progressMonitor;
		ModelMGR.getModelMGRInstance().suspendInfraFilesSaving();
		
		boolean suspendValidationNotification = false;
		if (BTValidator.instance().isNotificationEnabled()) {
			suspendValidationNotification = true;
			BTValidator.instance().setNotificationEnabled(false);
		}
		
		try {
			IImportOperation importOperation = null;	
			
			if (monitor != null) {			
				monitor.setTaskName(LoggingUtil.getLocalizedString(IeMessageKeys.TASK_STARTING_IMPORT, null));							
				monitor.beginTask(null, ProgressCalculator.getTotalProgress());
			}		
			
			try {
				if (projectName == null) {
					LoggingUtil.logError(IeMessageKeys.GENERIC_ERROR, IImportResultMessages.PROJECT_NAME_IS_NULL);
					LoggingUtil.traceExit(this, "execute");
					return;
				}
				if (operationId == null) {				
					LoggingUtil.logError(IeMessageKeys.GENERIC_ERROR, IImportResultMessages.OPERATION_ID_IS_NULL);
					LoggingUtil.traceExit(this, "execute");				
					return;
				}
				
				// if the queryCallback is null, we will handle any naming collision as ignore.			
				
				if (selectedFiles != null) {
					LoggingUtil.trace(this, "execute", "Selected files is not null.");
									
					ImportOperationRegistryReader reader = ImportOperationRegistryReader.getReader();
					importOperation = reader.findImportOperation(operationId);
					
					if (importOperation == null) {
						BTException e = new BTException(IeMessageKeys.GENERIC_ERROR);
						LoggingUtil.logError(IeMessageKeys.GENERIC_ERROR, IImportResultMessages.OPERATION_IS_NULL);					
						throw new InvocationTargetException(e);
					}
					LoggingUtil.trace(this, "execute", "Import Operation is loaded.");				
					
					// Create the import session				
					session = new ImportSession();				
					session.setProgressMonitor(monitor);
					session.setImportQuery(queryCallback);
					
					ImportOptions options = new ImportOptions();
					options.setAlwaysOverwrite(alwaysOverwrite);
					options.setAlwaysNewCopy(alwaysNewCopy);
					if(BLMImportExportConstants.xsdImportOperationId.equals(operationId)){
						options.setCollisionHandlingEnabled(true);						
					}
					if (additionalOptions != null) {
						LoggingUtil.trace(this, "execute", "Import Operation has additional options.");			
						options.setAdditionalOptions(additionalOptions);
					}
					session.setImportOption(options);								
					importOperation.setInputFiles(selectedFiles);
					importOperation.setProjectName(projectName);
					importOperation.setImportSession(session);
					
					// Begin
					if (monitor != null) {			
						monitor.worked(ProgressCalculator.getSimpleTask());
					}
					
					importOperation.readObjects();
					LoggingUtil.trace(this, "execute", "Import Operation is done.");	
					
					EList models = importOperation.getObjects();
					if (models == null || models.isEmpty()) {
						result = importOperation.getImportSession().getImportResult();
						//TODO AR
						ImportOperationRegistryReader.dispose();
						return;
					}
					
					EList predefinedTypes = importOperation.getPredefinedTypes();
					
					// Set the predefined types to the context.
					if (predefinedTypes.size() != 0) {
						session.getContext().put(IImportFramework.CONTEXT_PREDEFINED_REGISTRY_KEY, predefinedTypes);
						session.getContext().put(IImportFramework.CONTEXT_RESOLVED_TYPE, new ArrayList());
					}
					
					// Turn off the validation
					isEnabled = validator.isEnabled();
					validator.setEnabled(false);
					DependencyManager.instance().setEnabled(false);
								
					// The object is read
					if (monitor != null) {				
						monitor.worked(ProgressCalculator.getComplexTask());
					}							
								
					// The list of the models are all BOM root object with no
					// owningPackage. If the user pass in a non-root object or
					// root object with owningPackage, the object won't be 
					// handled.
					//
					// Currently this engine will just handle BOM object.
					// It is possible to extend to be a more generice engine
					// later.
					processor = new BOMProcessor();
					processor.setProjectName(projectName);
					processor.setImportSession(session);
					processor.setModels(models);
					processor.setIsRootModels(importOperation.isRootModels());
					processor.init();
					processor.process();	
					LoggingUtil.trace(this, "execute", "BOM Processor creates the models.");					
					
					// Update the BOM element stored in the ImportContext
					updateProcessor = new BOMUpdateProcessor();
					updateProcessor.setProjectName(projectName);
					updateProcessor.setImportSession(session);
					updateProcessor.process();
					LoggingUtil.trace(this, "execute", "BOM Update Processor updates the models.");				
					
					result = updateProcessor.getImportSession().getImportResult();

					//enable saving the master dependency model
					DependencyManager.instance().setEnabled(true);
					SaveDependencyModelCmd saveDMCmd = new SaveDependencyModelCmd();
					saveDMCmd.setProjectName(projectName);
					saveDMCmd.setProjectPath(FileMGR.getProjectPath(projectName));
					if(saveDMCmd.canExecute()){
						saveDMCmd.execute();	
					}
					// Turn on the validation
					validator.setEnabled(isEnabled);
					validateProject(projectName);
					
					// Handle Simulation import
					
					simProcessor = new BOMSimulationProcessor();
					simProcessor.setProjectName(projectName);
					simProcessor.setImportSession(session);
					simProcessor.setModels(models);
					if (monitor != null) {			
						monitor.setTaskName(LoggingUtil.getLocalizedString(IeMessageKeys.TASK_CREATING_SIMULATION_PROFILES, null));
					}
					simProcessor.process();
					
					LoggingUtil.trace(this, "execute", "BOM Simulation Processor created the simulation profile models.");				
				}
			} catch (OperationCanceledException e) {
				LoggingUtil.logWarning(IeMessageKeys.CANCEL_THE_OPERATION, null, e, IImportResultMessages.RUNTIME_EXCEPTION);			
				NavigatorUtil.save(projectName);	
				
				//enable saving the master dependency model
				DependencyManager.instance().setEnabled(true);
				SaveDependencyModelCmd saveDMCmd = new SaveDependencyModelCmd();
				saveDMCmd.setProjectName(projectName);
				saveDMCmd.setProjectPath(FileMGR.getProjectPath(projectName));
				if(saveDMCmd.canExecute()){
					saveDMCmd.execute();	
				}			
				
				// Turn on the validation
				validator.setEnabled(isEnabled);
				validateProject(projectName);
				//TODO AR	
				ImportOperationRegistryReader.dispose();	
				throw e; 
			} catch (RuntimeException e) {
				//defect 148904
				LoggingUtil.logError(session, IeMessageKeys.GENERIC_ERROR, null, e, IImportResultMessages.RUNTIME_EXCEPTION);
			//	LoggingUtil.logError(IeMessageKeys.GENERIC_ERROR, null, e, IImportResultMessages.RUNTIME_EXCEPTION);
				NavigatorUtil.save(projectName);
				// -- Debug --
				// e.printStackTrace();
				
				//enable saving the master dependency model
				DependencyManager.instance().setEnabled(true);
				SaveDependencyModelCmd saveDMCmd = new SaveDependencyModelCmd();
				saveDMCmd.setProjectName(projectName);
				saveDMCmd.setProjectPath(FileMGR.getProjectPath(projectName));
				if(saveDMCmd.canExecute()){
					saveDMCmd.execute();	
				}			
							
				// Turn on the validation
				validator.setEnabled(isEnabled);
				validateProject(projectName);
				//TODO AR	
				ImportOperationRegistryReader.dispose();
				throw new InvocationTargetException(e);
			} finally {
				// Clean up importOperation	
				if (importOperation != null) {
					importOperation.finish();
				}
				
				// The result of the engine need to be returned.
				if (updateProcessor != null) {
					LoggingUtil.trace(this, "execute", "The import engine updates the model.");	
					result = updateProcessor.getImportSession().getImportResult();
				} else if (processor != null) {
					LoggingUtil.trace(this, "execute", "The import engine only creates the model.");
					result = processor.getImportSession().getImportResult();	
				//AZ Add for getting the 1st step result messages
				}else if(importOperation != null){
					result = importOperation.getImportSession().getImportResult();
				}
				
				if (simProcessor != null) {
					LoggingUtil.trace(this, "execute", "The import engine created simulation profile model.");	
					result = simProcessor.getImportSession().getImportResult();
				}
				
				result = cleanseImportResult(result);
				
				if (result == null) {
					result = new ImportResult();
					result.addMessage(LoggingUtil.getMessage(IeMessageKeys.GENERIC_ERROR));			
				}
				
				if (monitor != null) {	
					monitor.done();
				}
				//TODO AR	
				ImportOperationRegistryReader.dispose();
			}
			
			
			ModelMGR.getModelMGRInstance().resumeInfraFilesSaving();
			ModelMGR.getModelMGRInstance().saveInfraFiles(projectName);
			//save the navigation nodes in the end as a whole to reduce foot print of the history copies
			NavigatorUtil.save(projectName);
			
			isInfraEnabled = true;

			if (isProjCleanUpReq == true) {
				// TN - 20060106: unload projects from memory in CR004281		
				String projectPath = FileMGR.getProjectPath(projectName);
				RemoveProjectCmd modelMgrCommand = new RemoveProjectCmd();
				modelMgrCommand.setProjectName(projectName);
				modelMgrCommand.setProjectPath(projectPath);
				if (modelMgrCommand.canExecute()) {
					modelMgrCommand.execute();
				}
				modelMgrCommand = null;
				
				// Unload project from File Manager.
				com.ibm.btools.blm.model.blmfilemanager.RemoveProjectCmd fileMgrCommand = new	
				com.ibm.btools.blm.model.blmfilemanager.RemoveProjectCmd();
				fileMgrCommand.setProjectName(projectName);
				if (fileMgrCommand.canExecute()) {
					fileMgrCommand.execute();
				}
				fileMgrCommand = null;		
				// End of TN - 20060106: unload projects from memory in CR004281
			}
			
			// Notify listeners
			if (suspendValidationNotification) {
				BTValidator.instance().setNotificationEnabled(true);
				BTValidator.instance().notifyListeners();
				suspendValidationNotification = false;
			}

			
			//System.out.println("END Try");
		} catch (InvocationTargetException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();

			// The following was added to support Excel Import
			if (additionalOptions != null
					&& additionalOptions.get("RETHROW_ALL_EXCEPTIONS") != null) {
				Boolean flag = (Boolean)additionalOptions.get("RETHROW_ALL_EXCEPTIONS");
				if(flag.booleanValue()) {
					throw e;
				}
			}
		}
		finally{
			if(!isInfraEnabled)
			{
				ModelMGR.getModelMGRInstance().resumeInfraFilesSaving();
				ModelMGR.getModelMGRInstance().saveInfraFiles(projectName);
				//save the navigation nodes in the end as a whole to reduce foot print of the history copies
				NavigatorUtil.save(projectName);
			}
			
			// Notify listeners
			if (suspendValidationNotification) {
				BTValidator.instance().setNotificationEnabled(true);
				BTValidator.instance().notifyListeners();
			}

			//System.out.println("Finally");
		}
		
		
		LoggingUtil.traceExit(this, "execute");
	}

	private void validateProject(String projectName) {
		if (projectName != null) {
			try {
				if (monitor != null) {			
					monitor.setTaskName(LoggingUtil.getLocalizedString(IeMessageKeys.TASK_VALIDATING_PROJECT, new String[] {projectName}));
				}				
				BOMUtil.validateProject(projectName);
				BTReporter.instance().saveActiveProjectMessages();
			} catch (ValidationException e) {
				LoggingUtil.trace(this, "execute", "The validation cannot be performed.");
				//re-throw runtime exception, so we don't do import simulation
				throw e;	
			}
			if (monitor != null) {			
				monitor.worked(ProgressCalculator.getSimpleTask());
			}				
		}
	}
	
	/**
	 * Get Import result
	 * @return the result of the import
	 */
	public ImportResult getImportResult() {
		return result;
	}

	/**
	 * Set always overwrite
	 * @param alwaysOverwrite
	 */
	public void setAlwaysOverwrite(final boolean alwaysOverwrite) {
		this.alwaysOverwrite = alwaysOverwrite;
	}

	/**
	 * Set always create a new copy
	 * @param alwaysNewCopy
	 */
	public void setAlwaysNewCopy(final boolean alwaysNewCopy) {
		this.alwaysNewCopy = alwaysNewCopy;
	}
	
	/**
	 * Set isProjCleanUpReq
	 * @param isProjCleanUpReq
	 */
	public void setIsProjCleanUpReq(final boolean isProjCleanUpReq) {
		this.isProjCleanUpReq = isProjCleanUpReq;
	}

	/**
	 * Get isProjCleanUpReq 
	 */
	public boolean getIsProjCleanUpReq() {
		return isProjCleanUpReq;
	}	
	
	/**
	 * This method is to remove the "PRE000102E" thrown from the platform
	 * @param result
	 * @return
	 */
	private ImportResult cleanseImportResult(ImportResult result){			
		ImportResult cleanedResult = new ImportResult();
		if(result == null){
			return null;
		}
		List messages = result.getMessages();
		if(messages == null || messages.isEmpty()){
			return result;
		}
		Iterator msgIt = messages.iterator();
		while(msgIt.hasNext()){
			Object each = msgIt.next();
			if(each instanceof ImportMessage){
				ImportMessage im = (ImportMessage)each;
				if(!(im.getErrorCode() != null && im.getErrorCode().startsWith("PRE"))){
					//we screen out the specific error codes					
					cleanedResult.addMessage(im);
				}
			}
		}
		return cleanedResult;
	}
}