package com.ibm.btools.blm.ie.imprt.registry;
/*
* Licensed Materials - Property of IBM
* 5724-I74, 5724-I75
*  Copyright IBM Corporation 2004, 2009. All rights reserved.
* US Government Users Restricted Rights - Use, duplication or disclosure
* restricted by GSA ADP Schedule Contract with IBM Corp.
*/
import java.util.HashMap;
import java.util.Map;

import org.eclipse.core.runtime.IConfigurationElement;
import org.eclipse.core.runtime.IExtension;
import org.eclipse.core.runtime.IExtensionPoint;
import org.eclipse.core.runtime.IExtensionRegistry;
import org.eclipse.core.runtime.Platform;
import org.osgi.framework.Bundle;

import com.ibm.btools.blm.ie.imprt.IImportFramework;
import com.ibm.btools.blm.ie.imprt.IImportOperation;
import com.ibm.btools.blm.ie.imprt.rule.util.LoggingUtil;

/**
 * Registry reader for the import operations	
 */
public class ImportOperationRegistryReader {
	// copyright	
	static final String copyright = "Licensed Material - Property of IBM " + " 5724-I74, 5724-I75 " + "(C) Copyright IBM Corporation 2004, 2009. All Rights Reserved. " + "U.S. Government Users Restricted Rights - Use, duplication or disclosure " + "restricted by GSA ADP Schedule Contract with IBM Corp.".intern(); //$NON-NLS-1$ //$NON-NLS-2$ //$NON-NLS-3$ //$NON-NLS-4$ //$NON-NLS-5$	

	protected Map importOperations;

	public static ImportOperationRegistryReader reader;
	
	/**
	 * Return the reader
	 * @return
	 */
	public static ImportOperationRegistryReader getReader() {
		if (reader == null) {
			// I didn't create a private constructor here becasue 
			// I give the application an optional to read the registry
			// everytime, so can save the memory if they don't want to 
			// use the static class
			reader = new ImportOperationRegistryReader();			
		}
		
		return reader;
	}

	/**
	 * The Empty map that contain the import operation and name pair
	 * @return
	 */
	protected Map createEmptyImportOperationCollection() {
		return new HashMap();
	}

	/**
	 * Read the Import Operations from the extensions
	 */	
	protected void readImportOperations() {
		LoggingUtil.traceEntry(this, "readImportOperations");		
		if(importOperations == null) {
			importOperations = createEmptyImportOperationCollection();
			// read the operations
			
			IExtensionRegistry pluginRegistry = Platform.getExtensionRegistry();
			IExtensionPoint point = null;
			IExtension[] extensions = null;
			IConfigurationElement[] elements = null;
			Object[] operationInfo = null;
			String className = null;
			String id = null;
			String name = null;
			ClassLoader classLoader = null;
			Class importOperationClass = null;
			IImportOperation importOperation = null;
			String elementName = null;
		
			// The extension point of the import operation is "com.ibm.btools.blm.ie.importOperations"
			point = pluginRegistry.getExtensionPoint(IImportFramework.IMPORT_OPERATION_EXTENSION_POINT);
			LoggingUtil.trace(this, "readImportOperations", "Get Extension Point from Plugin Registry.");			
			extensions = point.getExtensions();
			for (int i = 0; i < extensions.length; i++) {
				// Get the elements
				elements = extensions[i].getConfigurationElements();						
				for (int j = 0; j < elements.length; j++) {
					elementName = elements[j].getName();
					LoggingUtil.trace(this, "readImportOperations", "Get extension element: " + elementName + ".");					
					if (IImportFramework.IMPORT_OPERATION_ELEMENT_OPERATION.compareTo(elementName) == 0) {
						// Get the operation name by id
						id = elements[j].getAttribute(IImportFramework.IMPORT_OPERATION_ATT_ID);
						if (id == null) {
							LoggingUtil.trace(this, "readImportOperations", "The element id is null.");								
							break;
						}
						className = elements[j].getAttribute(IImportFramework.IMPORT_OPERATION_ATT_CLASS);
						if (className == null) {
							LoggingUtil.trace(this, "readImportOperations", "The element class name is null.");							
							break;
						}						
						name = elements[j].getAttribute(IImportFramework.IMPORT_OPERATION_ATT_NAME);
						if (name == null) {
							// Name is required but we can change the schema to support it as optional
							// If this is the case, we just use the id as the name.
							LoggingUtil.trace(this, "readImportOperations", "The element name is null.");							
							name = id;						
						}
					
						// Load the class
						Bundle plugin = Platform.getBundle(extensions[i].getNamespace());
						try {						
							importOperationClass = plugin.loadClass(className);
							importOperation = (IImportOperation) importOperationClass.newInstance();
							importOperation.setName(name);
						} catch (Exception e) {
							// skip
							LoggingUtil.trace(this, "readImportOperations", "Cannot load the element class.");							
							break;						
						}
						operationInfo = new Object[] {name, importOperation};
						
						importOperations.put(id, operationInfo);						
					}		
				}					
			}
		}
		LoggingUtil.traceExit(this, "readImportOperations");		
	}	
	
	/**
	 * Return a map of the import operations.
	 * @return Map of import operations
	 */	
	public Map getImportOperations() {
		if(importOperations == null)
			readImportOperations();
		return importOperations;
	}
		
	/**
	 * Return the import operation by id.
	 * @param id
	 * @return
	 */		
	public IImportOperation findImportOperation(final String id) {
		LoggingUtil.traceEntry(this, "findImportOperation");		
		Map importOperations = getImportOperations();
		Object[] operationInfo = (Object[]) importOperations.get(id);
		if (operationInfo == null) {
			LoggingUtil.trace(this, "findImportOperation", "Cannot find the operation: " + id + ".");			
			LoggingUtil.traceExit(this, "findImportOperation");			
			return null;
		}
		
		Object operation = operationInfo[1];
		if (operation != null) {		
		
			if (operation instanceof IImportOperation) {				
				LoggingUtil.traceExit(this, "findImportOperation");				
				return (IImportOperation) operation; 
			}
			LoggingUtil.trace(this, "findImportOperation", "The operation is not a proper class: " + operation.getClass() + ".");			
		}
		
		LoggingUtil.traceExit(this, "findImportOperation");		
		return null;
	}	
	
	public static void dispose(){
		reader = null;
		
	}
}
