package com.ibm.btools.blm.ie.imprt.registry;
/*
* Licensed Materials - Property of IBM
* 5724-I74, 5724-I75
*  Copyright IBM Corporation 2004, 2009. All rights reserved.
* US Government Users Restricted Rights - Use, duplication or disclosure
* restricted by GSA ADP Schedule Contract with IBM Corp.
*/

import java.util.HashMap;
import java.util.Map;

import org.eclipse.core.runtime.IConfigurationElement;
import org.eclipse.core.runtime.IExtension;
import org.eclipse.core.runtime.IExtensionPoint;
import org.eclipse.core.runtime.IExtensionRegistry;
import org.eclipse.core.runtime.Platform;
import org.osgi.framework.Bundle;

import com.ibm.btools.blm.ie.imprt.IImportFramework;
import com.ibm.btools.blm.ie.imprt.rule.IImportRule;
import com.ibm.btools.blm.ie.imprt.rule.util.LoggingUtil;

/**
 * Registry reader for the import rules	
 */
public class ImportRuleRegistryReader {
	// copyright	
	static final String copyright = "Licensed Material - Property of IBM " + " 5724-I74, 5724-I75 " + "(C) Copyright IBM Corporation 2004, 2009. All Rights Reserved. " + "U.S. Government Users Restricted Rights - Use, duplication or disclosure " + "restricted by GSA ADP Schedule Contract with IBM Corp.".intern(); //$NON-NLS-1$ //$NON-NLS-2$ //$NON-NLS-3$ //$NON-NLS-4$ //$NON-NLS-5$	

	protected Map importRules;

	public static ImportRuleRegistryReader reader;
	
	/**
	 * Return the reader
	 * @return
	 */
	public static ImportRuleRegistryReader getReader() {
		if (reader == null) {
			// I didn't create a private constructor here becasue 
			// I give the application an optional to read the registry
			// everytime, so can save the memory if they don't want to 
			// use the static class
			reader = new ImportRuleRegistryReader();			
		}
		
		return reader;
	}

	/**
	 * The Empty map that contain the import rule and name pair
	 * @return
	 */
	protected Map createEmptyImportRuleCollection() {
		return new HashMap();
	}

	/**
	 * Read the Import Rules from the extensions
	 */	
	protected void readImportRules() {
		LoggingUtil.traceEntry(this, "readImportOperations");		
		if(importRules == null) {
			importRules = createEmptyImportRuleCollection();
			// read the rules
			
			IExtensionRegistry pluginRegistry = Platform.getExtensionRegistry();
			IExtensionPoint point = null;
			IExtension[] extensions = null;
			IConfigurationElement[] elements = null;
			Object[] ruleInfo = null;
			String className = null;
			String name = null;
			String sequence;
			ClassLoader classLoader = null;
			Class importRuleClass = null;
			IImportRule importRule = null;
			Object sequenceTable = null;
			Object rules = null;
			String elementName = null;
		
			// The extension point of the import rule is "com.ibm.btools.blm.ie.importRules"
			point = pluginRegistry.getExtensionPoint(IImportFramework.IMPORT_RULE_EXTENSION_POINT);
			LoggingUtil.trace(this, "readImportOperations", "Get Extension Point from Plugin Registry.");			
			extensions = point.getExtensions();
			for (int i = 0; i < extensions.length; i++) {
				// Get the elements
				elements = extensions[i].getConfigurationElements();						
				for (int j = 0; j < elements.length; j++) {
					elementName = elements[j].getName();
					LoggingUtil.trace(this, "readImportOperations", "Get extension element: " + elementName + ".");					
					if (elementName != null && IImportFramework.IMPORT_RULE_ELEMENT_RULE.compareTo(elementName) == 0) {

						className = elements[j].getAttribute(IImportFramework.IMPORT_RULE_ATT_CLASS);
						if (className == null) {
							LoggingUtil.trace(this, "readImportOperations", "The element class name is null.");							
							break;
						}						
						name = elements[j].getAttribute(IImportFramework.IMPORT_RULE_ATT_NAME);
						if (name == null) {
							LoggingUtil.trace(this, "readImportOperations", "The element name is null.");							
							break;						
						}
						sequence = elements[j].getAttribute(IImportFramework.IMPORT_RULE_ATT_SEQUENCE);
						if (sequence == null) {
							LoggingUtil.trace(this, "readImportOperations", "The element sequence is null.");							
							break;						
						}	
						try {
							new Integer(sequence);
						} catch (NumberFormatException e) {
							LoggingUtil.trace(this, "readImportOperations", "The element sequence is not an integer.");							
							break;
						}
					
						// Load the class
						Bundle plugin = Platform.getBundle(extensions[i].getNamespace());
						try {						
							importRuleClass = plugin.loadClass(className);
							importRule = (IImportRule) importRuleClass.newInstance();
						} catch (Exception e) {
							// skip
							LoggingUtil.trace(this, "readImportOperations", "Cannot load the element class.");							
							break;						
						}
					
						sequenceTable = importRules.get(sequence);
						if (sequenceTable == null || !(sequenceTable instanceof Map)) {
							sequenceTable = new HashMap();
							importRules.put(sequence, sequenceTable);
						}
						
						rules = ((Map)sequenceTable).get(name);
						if (rules == null || !(rules instanceof Map)) {
							rules = new HashMap();
							((Map)sequenceTable).put(name, rules);
						}
						((Map)rules).put(name, importRule);
						
						importRules.put(sequence, sequenceTable);						
					}		
				}					
			}
		}
		LoggingUtil.traceExit(this, "readImportOperations");		
	}	
	
	/**
	 * Return a map of the import rules.
	 * @return Map of import rules
	 */	
	public Map getImportRules() {
		if(importRules == null)
			readImportRules();
		return importRules;
	}
	
	/**
	 * Return the import rule by sequence and type
	 * @param sequence
	 * @param name
	 * @return
	 */
	public IImportRule getRule(final String sequence, final String type) {
		LoggingUtil.traceEntry(this, "getRule");		
		Map sequences = getImportRules();
		Object sequenceTable = sequences.get(sequence);
		if (sequenceTable == null || !(sequenceTable instanceof Map)) {
			LoggingUtil.trace(this, "getRule", "The sequence table is not created.");			
			LoggingUtil.traceExit(this, "getRule");			
			return null;
		}
		Object rules = ((Map)sequenceTable).get(type);
		if (rules == null || !(rules instanceof Map)) {
			LoggingUtil.trace(this, "getRule", "The rule table is not created.");
			LoggingUtil.traceExit(this, "getRule");			
			return null;
		}
		
		Object rule =((Map)rules).get(type);	
		
		if (rule != null) {
			if (rule instanceof IImportRule) {			
				LoggingUtil.traceExit(this, "getRule");			
				return (IImportRule) rule;
			} else {			
				LoggingUtil.trace(this, "findImportOperation", "The rule is not a proper class: " + rule.getClass() + ".");

			}
		} 
		
		LoggingUtil.traceExit(this, "getRule");
		return null;		
	}
	
	public static void dispose(){
		reader = null;
		
	}
}
