package com.ibm.btools.te.ilm.heuristics.abstractbpel.util;
/*
* Licensed Materials - Property of IBM
* 5724-I74, 5724-I75
*  Copyright IBM Corporation 2003, 2009. All rights reserved.
* US Government Users Restricted Rights - Use, duplication or disclosure
* restricted by GSA ADP Schedule Contract with IBM Corp.
*/

import java.util.HashMap;
import java.util.Iterator;
import java.util.Map;

import org.eclipse.xsd.XSDElementDeclaration;
import org.eclipse.xsd.XSDSimpleTypeDefinition;
import org.eclipse.xsd.XSDTypeDefinition;
import org.eclipse.xsd.util.XSDConstants;
import org.eclipse.xsd.util.XSDUtil;

import com.ibm.btools.bom.model.artifacts.NamedElement;
import com.ibm.btools.bom.model.processes.activities.Action;
import com.ibm.btools.bom.model.processes.activities.InputPinSet;
import com.ibm.btools.bom.model.processes.activities.OutputPinSet;
import com.ibm.btools.bom.model.processes.activities.PinSet;
import com.ibm.btools.te.framework.TransformationContext;
import com.ibm.btools.te.ilm.heuristics.helper.BomUtils;
import com.ibm.btools.te.ilm.heuristics.naming.JavaNCNameConverter;
import com.ibm.btools.te.ilm.heuristics.naming.NamingRegistry;
import com.ibm.wbit.bpel.Activity;
import com.ibm.wbit.bpel.Assign;
import com.ibm.wbit.bpel.BPELFactory;
import com.ibm.wbit.bpel.BPELVariable;
import com.ibm.wbit.bpel.Copy;
import com.ibm.wbit.bpel.From;
import com.ibm.wbit.bpel.To;
import com.ibm.wbit.bpelpp.BPELPlusFactory;
import com.ibm.wbit.bpelpp.False;
import com.ibm.wbit.bpelpp.True;

/**
 * This util class maintains a registry of the mapping of SCOPE start and end activities and variables to 
 * 	Keys: BOM NamedElements (StructuredActivityNode, LoopNode, InputPinSet, and OutputPinSet)
 *  Values: LocalContainerPinSetsEntry (Activity and Variable)
 * - START activities:
 * --- Empty for StructuredActivityNode, LoopNode with a single regular OutputPinSet
 * --- Assign for StructuredActivityNode, LoopNode with multiple regular OutputPinSets
 * - END activities:
 * --- Throw for exception OutputPinSet
 * --- Assign for an OutputPinSet in Action with multiple regular OutputPinSets
 */
public class ScopeRegistryUtil {

	public static final String copyright = "Licensed Material - Property of IBM " + " 5724-I74, 5724-I75 " + "(C) Copyright IBM Corporation 2003, 2009. All Rights Reserved. " + "U.S. Government Users Restricted Rights - Use, duplication or disclosure " + "restricted by GSA ADP Schedule Contract with IBM Corp.".intern(); //$NON-NLS-1$ //$NON-NLS-2$ //$NON-NLS-3$ //$NON-NLS-4$ //$NON-NLS-5$
	
	public static String REGISTRY_KEY = "ScopeRegistry";
	public static String VARIABLE = "Variable";
	public static String RETURN_STATEMENT_START = "return (";
	public static String RETURN_STATEMENT_END = ".booleanValue() == true);";
	
	
	public static class LocalContainerPinSetsEntry {
		private Activity activity = null;
		private BPELVariable variable = null;
		
		public LocalContainerPinSetsEntry() {
			super();
		}
		public void setActivity(Activity activity) {
			this.activity = activity;
		}
		public Activity getActivity() {
			return activity;
		}
		public void setVariable(BPELVariable variable) {
			this.variable = variable;
		}
		public BPELVariable getVariable() {
			return variable;
		}
	}

	public static Map <NamedElement, LocalContainerPinSetsEntry> getRegistry(TransformationContext context) {
		Map <NamedElement, LocalContainerPinSetsEntry>registry = (Map)context.get(REGISTRY_KEY);
		if (registry==null) {
			registry = new HashMap<NamedElement, LocalContainerPinSetsEntry>();
			context.put(REGISTRY_KEY, registry);
		}
		return registry;
	}
	
	/**
	 * Register BOM NamedElement with the bpel activity
	 * Warning - may overwrite existing activity
	 * @param context
	 * @param namedElement
	 * @param bpelAct
	 */
	public static void registerNamedElement(TransformationContext context, NamedElement namedElement, Activity bpelAct) {
		Map <NamedElement, LocalContainerPinSetsEntry> map = getRegistry(context);
		if (map.containsKey(namedElement)) {
			LocalContainerPinSetsEntry localContainer = map.get(namedElement);
			localContainer.setActivity(bpelAct);
		} else{
			LocalContainerPinSetsEntry localContainer = new LocalContainerPinSetsEntry();
			localContainer.setActivity(bpelAct);
			map.put(namedElement, localContainer);
		}
	}

	/**
	 * Register BOM NamedElement with the bpel variable
	 * Warning - may overwrite existing activity
	 * @param context
	 * @param namedElement
	 * @param bpelAct
	 */
	public static void registerNamedElement(TransformationContext context, NamedElement namedElement, BPELVariable bpelVar) {
		Map <NamedElement, LocalContainerPinSetsEntry> map = getRegistry(context);
		if (map.containsKey(namedElement)) {
			LocalContainerPinSetsEntry localContainer = map.get(namedElement);
			localContainer.setVariable(bpelVar);
		} else{
			LocalContainerPinSetsEntry localContainer = new LocalContainerPinSetsEntry();
			localContainer.setVariable(bpelVar);
			map.put(namedElement, localContainer);
		}
	}

	/**
	 * get the bpel activity from the context
	 * @param context
	 * @param namedElement
	 * return Activity
	 */
	public static Activity getBpelActivity(TransformationContext context, NamedElement namedElement) {
		Map <NamedElement, LocalContainerPinSetsEntry> map = getRegistry(context);
		if (map.containsKey(namedElement)) {
			LocalContainerPinSetsEntry localContainer = map.get(namedElement);
			return localContainer.getActivity();
		}
		return null;
	}
	
	/**
	 * get the bpel variable from the context
	 * @param context
	 * @param namedElement
	 * return Variable
	 */
	public static BPELVariable getBpelVariable(TransformationContext context, NamedElement namedElement) {
		Map <NamedElement, LocalContainerPinSetsEntry> map = getRegistry(context);
		if (map.containsKey(namedElement)) {
			LocalContainerPinSetsEntry localContainer = map.get(namedElement);
			return localContainer.getVariable();
		}
		return null;
	}
	
	/**
	 * checks whether variable exists for this named element
	 * @param context
	 * @param namedElement
	 * return boolean
	 */
	public static boolean hasVariable(TransformationContext context, NamedElement namedElement) {
		Map <NamedElement, LocalContainerPinSetsEntry> map = getRegistry(context);
		if (map.containsKey(namedElement)) {
			LocalContainerPinSetsEntry localContainer = map.get(namedElement);
			if (localContainer.getVariable() != null) {
				return true;
			}
		}
		return false;
	}

	
	/**
	 * Creates Copy block for assign: 
	 * 		- boolean "true" to variable representing the given pin set of local container (SAN, LoopNode)
	 * 		- boolean "false" to all other variables of either Input or Output pin set of local container (SAN, LoopNode)
	 * @param assignAct Assign
	 * @param pinSet PinSet
	 * @param context TransformationContext
	 * return 
	 */
	public static void createCopyForPinSetsINContainer(Assign assignAct, PinSet pinSet, TransformationContext context) {
		if (assignAct == null || pinSet == null) {
			return;
		}
		
		if (pinSet instanceof InputPinSet) {
			InputPinSet thisISet = (InputPinSet)pinSet;
			Iterator <OutputPinSet> it = thisISet.getAction().getInputPinSet().iterator();
			InputPinSet iSet = null;
			while (it.hasNext()) {
				iSet = (InputPinSet)it.next();
				com.ibm.wbit.bpel.Variable var = ScopeRegistryUtil.getBpelVariable(context, iSet);
				if (var != null) { //should be created already
					Copy copyForAssign = BPELFactory.eINSTANCE.createCopy();
					assignAct.getCopy().add(copyForAssign);					
					From fromValue = BPELFactory.eINSTANCE.createFrom();
					copyForAssign.setFrom(fromValue);
					if (iSet == pinSet) {
						//TODO change the deprecated API into setLieratlElement("true") post 6.2
						fromValue.setLiteral("true");	
						//defect 81875 get rid of the CDATA tag for the literal
						fromValue.setUnsafeLiteral(Boolean.FALSE);
					} else {
						fromValue.setLiteral("false");		
						fromValue.setUnsafeLiteral(Boolean.FALSE);
					}
					To toValue = BPELFactory.eINSTANCE.createTo();
					copyForAssign.setTo(toValue);
					toValue.setVariable(var);
				}
			}			
		} else if (pinSet instanceof OutputPinSet) {
			OutputPinSet thisOSet = (OutputPinSet)pinSet;
			Iterator <OutputPinSet> it = thisOSet.getAction().getOutputPinSet().iterator();
			OutputPinSet oSet = null;
			while (it.hasNext()) {
				oSet = (OutputPinSet)it.next();
				if ((oSet.getIsException() != null && oSet.getIsException().booleanValue()) ||
						(oSet.getIsStream() != null && oSet.getIsStream().booleanValue())) {
					continue;
				}
				com.ibm.wbit.bpel.BPELVariable var = ScopeRegistryUtil.getBpelVariable(context, oSet);
				if (var != null) { //should be created already
					Copy copyForAssign = BPELFactory.eINSTANCE.createCopy();
					assignAct.getCopy().add(copyForAssign);
			
					From fromValue = BPELFactory.eINSTANCE.createFrom();
					copyForAssign.setFrom(fromValue);
					if (oSet == pinSet) {
						fromValue.setLiteral("true");
						fromValue.setUnsafeLiteral(Boolean.FALSE);;
					} else {
						fromValue.setLiteral("false");
						fromValue.setUnsafeLiteral(Boolean.FALSE);
					}
					To toValue = BPELFactory.eINSTANCE.createTo();
					copyForAssign.setTo(toValue);
					toValue.setVariable(var);
				}
			}			
		}
	}
	
	/**
	 * Creates Copy block for assign from boolean "true" to variable representing pin set of local container (SAN, LoopNode)
	 * @param assignAct Assign
	 * @param pinSet PinSet
	 * @param context TransformationContext
	 * @param nameRegistry NamingRegistry
	 * @param converter JavaNCNameConverter
	 * 
	 */
	public static void createTrueBooleanAssignForPinSetINContainer(Assign assignAct, PinSet ps, TransformationContext context, NamingRegistry nameRegistry, JavaNCNameConverter converter) {
		if (assignAct == null || ps == null) {
			return;
			
		}
		BPELVariable var = ScopeRegistryUtil.getBpelVariable(context, ps);
		if (var == null) { //should be created already
			var = createBooleanVariable(ps, context, nameRegistry, converter);
		}
		
		Copy copyForAssign = BPELFactory.eINSTANCE.createCopy();
		assignAct.getCopy().add(copyForAssign);

		From fromValue = BPELFactory.eINSTANCE.createFrom();
		copyForAssign.setFrom(fromValue);
		fromValue.setLiteral("true");
		fromValue.setUnsafeLiteral(Boolean.FALSE);
		To toValue = BPELFactory.eINSTANCE.createTo();
		copyForAssign.setTo(toValue);
		toValue.setVariable(var);
	}
	

	/**
	 * Creates boolean variables that will be used for entry/exit set for local process and loop
	 * @param pinSet PinSet
	 * @param context TransformationContext
	 * @param nameRegistry NamingRegistry
	 * @param converter JavaNCNameConverter
	 */
	public static BPELVariable createBooleanVariable(PinSet pinSet, TransformationContext context, NamingRegistry nameRegistry, JavaNCNameConverter converter){
		if (pinSet == null) {
			return null;
		}
		BPELVariable newVariable = BPELFactory.eINSTANCE.createBPELVariable();
		String pinSetName = BomUtils.getCanonicalName(pinSet);
		String convertedName = converter.convertName(nameRegistry, newVariable, pinSetName);

		newVariable.setName(convertedName + VARIABLE);
		XSDSimpleTypeDefinition xsdType = XSDUtil.getSchemaForSchema(XSDConstants.SCHEMA_FOR_SCHEMA_URI_2001).resolveSimpleTypeDefinition("boolean");
		newVariable.setType(xsdType);
		
		ScopeRegistryUtil.registerNamedElement(context, pinSet, newVariable);
		return newVariable;
	}
		
	/**
	 * Creates variables that will be used for entry/exit set for local process and loop
	 * @param pinSet OutputPinSet
	 * @param context TransformationContext
	 * @param nameRegistry NamingRegistry
	 * @param xsdElement XSDElementDeclaration
	 * @param converter JavaNCNameConverter
	 * 
	 */
	public static BPELVariable createVariableForExceptionPinSet(OutputPinSet pinSet, 
									TransformationContext context, 
									NamingRegistry nameRegistry, 
									XSDElementDeclaration xsdElement,
									JavaNCNameConverter converter){
		if (pinSet == null) {
			return null;
		}
		BPELVariable newVariable = BPELFactory.eINSTANCE.createBPELVariable();
		String pinSetName = BomUtils.getCanonicalName(pinSet);
		String convertedName = converter.convertName(nameRegistry, newVariable, pinSetName);

		newVariable.setName(convertedName + VARIABLE);
		newVariable.setXSDElement(xsdElement);
		
		return newVariable;
	}

	/**
	 * Creates variables that will be used for entry/exit set for local process and loop
	 * @param pinSet OutputPinSet
	 * @param context TransformationContext
	 * @param nameRegistry NamingRegistry
	 * @param xsdElement XSDElementDeclaration
	 * @param converter JavaNCNameConverter
	 * 
	 */
	public static BPELVariable createVariableForExceptionPinSet(OutputPinSet pinSet, 
									TransformationContext context, 
									NamingRegistry nameRegistry, 
									XSDTypeDefinition xsdType,
									JavaNCNameConverter converter){
		if (pinSet == null) {
			return null;
		}
		BPELVariable newVariable = BPELFactory.eINSTANCE.createBPELVariable();
		String pinSetName = BomUtils.getCanonicalName(pinSet);
		String convertedName = converter.convertName(nameRegistry, newVariable, pinSetName);

		newVariable.setName(convertedName + VARIABLE);
		newVariable.setType(xsdType);
		
		ScopeRegistryUtil.registerNamedElement(context, pinSet, newVariable);
		return newVariable;
	}

	/**
	 * Creates Copy block for assign from boolean "true" to variable representing pin set of local container (SAN, LoopNode)
	 * @param assignAct Assign
	 * @param action Action
	 * @param context TransformationContext
	 */
	public static void createFalseBooleanAssignForOutputPinSetsINContainer(Assign assignAct, Action action, TransformationContext context) {
		if (assignAct == null || action == null || 
				action.getOutputPinSet() == null) {
			return;
		}
		
		Iterator <OutputPinSet> it = action.getOutputPinSet().iterator();
		OutputPinSet oSet = null;
		while (it.hasNext()) {
			oSet = (OutputPinSet)it.next();
			if ((oSet.getIsException() != null && oSet.getIsException().booleanValue()) ||
					(oSet.getIsStream() != null && oSet.getIsStream().booleanValue())) {
				continue;
			}
			BPELVariable var = ScopeRegistryUtil.getBpelVariable(context, oSet);
			if (var != null) { //should be created already
				Copy copyForAssign = BPELFactory.eINSTANCE.createCopy();
				assignAct.getCopy().add(copyForAssign);
		
				From fromValue = BPELFactory.eINSTANCE.createFrom();
				copyForAssign.setFrom(fromValue);
				fromValue.setLiteral("false");
				fromValue.setUnsafeLiteral(Boolean.FALSE);
				To toValue = BPELFactory.eINSTANCE.createTo();
				copyForAssign.setTo(toValue);
				toValue.setVariable(var);
			}
		}
	}

	
	public static boolean hasExceptionWithMultiPins(Action action) {
		if (action == null || action.getOutputPinSet() == null || action.getOutputPinSet().isEmpty()) {
			return false;
		}
		boolean hasMultiplePinInException = false;
		Iterator <OutputPinSet> it = action.getOutputPinSet().iterator();
		OutputPinSet oSet = null;
		while (it.hasNext() && !hasMultiplePinInException) {
			oSet = (OutputPinSet)it.next();
			if (oSet.getIsException() != null && oSet.getIsException().booleanValue()) {
				if (oSet.getOutputObjectPin().size() > 1) {
					hasMultiplePinInException = true;
				}
			} else {
				continue;
			}
		}			
		return hasMultiplePinInException;
	}
}