/**
 * <copyright>
 *
 *    Copyright IBM Corporation 2003, 2009. All rights reserved.
 *
 * </copyright>
 *
 * %W%
 * @version %I% %H%
 */
package com.ibm.btools.te.ilm.heuristics.bpelp.util;
/*
* Licensed Materials - Property of IBM
* 5724-I74, 5724-I75
*  Copyright IBM Corporation 2008, 2009. All rights reserved.
* US Government Users Restricted Rights - Use, duplication or disclosure
* restricted by GSA ADP Schedule Contract with IBM Corp.
*/

import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;

import org.eclipse.emf.ecore.EObject;

import com.ibm.btools.bom.model.artifacts.LiteralSpecification;
import com.ibm.btools.bom.model.processes.activities.ObjectPin;
import com.ibm.btools.expression.bom.model.StructuredOpaqueExpression;
import com.ibm.btools.expression.function.CoreFunctions;
import com.ibm.btools.expression.model.Expression;
import com.ibm.btools.expression.model.FunctionArgument;
import com.ibm.btools.expression.model.FunctionExpression;
import com.ibm.btools.expression.model.ModelPathExpression;
import com.ibm.btools.expression.model.ReferenceStep;
import com.ibm.btools.expression.xpath.XPathExpressionSerializer;
import com.ibm.btools.te.ilm.heuristics.abstractbpel.impl.MapRuleImpl;
import com.ibm.wbit.bpel.BPELFactory;
import com.ibm.wbit.bpel.BPELVariable;
import com.ibm.wbit.bpel.Copy;
import com.ibm.wbit.bpel.From;
import com.ibm.wbit.bpel.Query;
import com.ibm.wbit.bpel.To;

/**
 * Instance of this class responsible for analyze and keep track of the copies. When a new copy is added, determine if this copy create a value swapping
 * scenario (A.X -> A.Y, A.Y -> A.X), if so, update the copies and obtain the temp variable for value swapping (A.X -> temp, A.Y -> A.X, temp -> A.Y). 
 *
 */
public class SwapOrganizer {
	public static final String copyright = "Licensed Material - Property of IBM " + " 5724-I74, 5724-I75 " + "(C) Copyright IBM Corporation 2003, 2009. All Rights Reserved. " + "U.S. Government Users Restricted Rights - Use, duplication or disclosure " + "restricted by GSA ADP Schedule Contract with IBM Corp.".intern(); //$NON-NLS-1$ //$NON-NLS-2$ //$NON-NLS-3$ //$NON-NLS-4$ //$NON-NLS-5$
	// the rule that instantiated this manager, use this reference to access the contents of the rule
	private MapRuleImpl mapRule = null;
	// the list of copies/records were currently kept in the manager, they provides APIs help to identify swapping scenario with the new incoming copy/record
	private List<CopyRecord> copyRecords = new ArrayList<CopyRecord>();
	// the map contains the mapping between pins and bpel variables.
	private java.util.Map<ObjectPin, BPELVariable> pin2var = null;
	// the list of copies/records created as a result of a collision split (tempVariable.Att1 -> variable.att1)
	private List<CopyRecord> splitRecords = new ArrayList<CopyRecord>();
	
	/**
	 * Constructor
	 * @param mRule
	 */
	public SwapOrganizer(MapRuleImpl mRule, java.util.Map<ObjectPin, BPELVariable> pinToVar){
		super();
		this.mapRule = mRule;
		this.pin2var = pinToVar;
	}
	
	/**
	 * Append a copy to the current list of copies. The appended copy will be analyzed against the existing copies to identified collision.
	 * A collision means the incoming source exists as the previous copy's target, i.e. this indicates the incoming source would have already been modified 
	 * in previous copy. To compensate, the previous target will need to be 'swapped' to a temp variable so that the original source value would not be lost. 
	 * @param aCopy
	 */
	public void append(Copy aCopy){
		CopyRecord newRecord = new CopyRecord(aCopy);
//		if (! newRecord.isPassThrough()){
			for (CopyRecord existingRecord: copyRecords){
				renderRecord(newRecord, existingRecord);
			}
//		}
		copyRecords.add(newRecord);
	}
	
	/**
	 * Check each incoming fromKeys against the existingToKey. If the existingToKey match (or is parent of) any incoming fromKeys
	 * then return true, otherwise return false. Return true means the current source (or one of the sources in case of concat) of 
	 * the copy is the target of the previous/existing copy.
	 * @param fromKeys
	 * @param existingToKey
	 * @return
	 */
	private boolean isSourceModified(List<String> fromKeys, String existingToKey){
		if (fromKeys.contains(existingToKey)){
			return true;
		}
		boolean result = false;
		for (String aFromKey: fromKeys){
			if (aFromKey.startsWith(existingToKey)){
				int index = existingToKey.length();
				if (index < aFromKey.length() && aFromKey.substring(index, index).equals(XPathExpressionSerializer.STEP_SEP)){				
					result = true;
					break;
				}
			}
		}
		return result;
	}
	
	/**
	 * Check the incoming toKey against the existing toKey. If the existing toKey match (or is parent of) the incoming toKey,
	 * then return true, otherwise return false. Return true means the current target of the copy is the target of the previous/
	 * existing copy.
	 * @param incomingToKey
	 * @param existingToKey
	 * @return
	 */
	private boolean isTargetModified(String incomingToKey, String existingToKey){
		if (incomingToKey.equals(existingToKey)){
			return true;
		}
		boolean result = false;
		if (incomingToKey.startsWith(existingToKey)){
			int index = existingToKey.length();
			if (index < incomingToKey.length() && incomingToKey.substring(index, index).equals(XPathExpressionSerializer.STEP_SEP)){
				result = true;
			}
		}
		return result;
	}

	private void renderRecord(CopyRecord newRecord, CopyRecord existingRecord){
		//first, check if the new record represents a pass-through (i.e. assigning A to A or a.b to a.b etc). If so, skip it as it is not part of the swapping.
		if (existingRecord.isPassThrough()){
			return;
		}
		//second, check incoming source against existing target, if collision was found, swap the existing target variable to use the temp variable.
		if (isSourceModified(newRecord.getFromKeys(), existingRecord.getToKey())){
			if (! existingRecord.isSwapped()){
				existingRecord.swapTarget();
				// create the corresponding copy that assign the original value back from temp variable to original variable, these copies will be appended to the end of the existing.
				// note: 	Duplication will needed to be checked
				//			New copy needed to be registered in MapRule's copy map with the corresponding SOE of the original copy.
				CopyRecord splitRecord = createSplitRecord(existingRecord);
				if (splitRecord != null)
					registerSplitRecord(splitRecord);
			}
		}
		//third, check incoming target against existing target, if collision was found swap the incoming target to use the temp variable IF the existing target was swapped.
		if (! newRecord.isSwapped() && isTargetModified(newRecord.getToKey(), existingRecord.getToKey())){
			if (existingRecord.isSwapped()){
				newRecord.swapTarget();
				// No need to create corresponding copy here, since the same one should have been created already when the existing record was swapped.
			}
		}
	}
	
	/**
	 * Register this split record with the same expression of the referenced copy. 
	 * This is for traceability support.
	 * @param splitRecord
	 */
	private void registerSplitRecord(CopyRecord splitRecord){
		Copy reference = splitRecord.getReference();
		Copy copy = splitRecord.getCopy();
		EObject value = mapRule.getObjectMappedToCopy(reference);
		if (value != null){
			mapRule.register(copy, value);
		}
		splitRecords.add(splitRecord);
	}
	
	private CopyRecord createSplitRecord(CopyRecord swappedRecord){
		CopyRecord result = null;
		
		Copy swappedCopy = swappedRecord.getCopy();
		BPELVariable tempVar = (BPELVariable)swappedCopy.getTo().getVariable();
		BPELVariable var = mapRule.getOriginalVariable(tempVar);
		String toKey = swappedRecord.getToKey();
		int index = toKey.indexOf(XPathExpressionSerializer.STEP_SEP);
		String path = index == -1 ? "" : toKey.substring(index + 1);
		
		// check duplication
		if (isSplitRecordExist(tempVar, var, path)){
			return result;
		}
		
		Copy copy = BPELFactory.eINSTANCE.createCopy();
		
		From from = BPELFactory.eINSTANCE.createFrom();
		from.setVariable(tempVar);
		if (path != null && ! "".equals(path)){
			Query queryFrom = BPELFactory.eINSTANCE.createQuery();
			queryFrom.setQueryLanguage(MapRuleImpl.QUERY_LANGUAGE);
			queryFrom.setValue(path);
			from.setQuery(queryFrom);
		}
		
		To to = BPELFactory.eINSTANCE.createTo();		
		to.setVariable(var);
		if (path != null && ! "".equals(path)){
			Query queryTo = BPELFactory.eINSTANCE.createQuery();
			queryTo.setQueryLanguage(MapRuleImpl.QUERY_LANGUAGE);
			queryTo.setValue(path);
			to.setQuery(queryTo);
		}
		
		copy.setFrom(from);
		copy.setTo(to);
		
		result = new CopyRecord(copy);
		result.setReference(swappedCopy);
		
		return result;
	}
	
	private boolean isSplitRecordExist(BPELVariable fromVar, BPELVariable toVar, String path){
		boolean result = false;
		for (CopyRecord r: splitRecords){
			Copy c = r.getCopy();
			if (c.getFrom().getVariable() == fromVar && c.getTo().getVariable() == toVar && 
				c.getFrom().getQuery().getValue().equals(path) && c.getTo().getQuery().getValue().equals(path)){
				result = true;
				break;
			}				
		}
		return result;
	}
	
	public List<Copy> getRenderedCopies(){
		List<Copy> result = new ArrayList<Copy>();
		for (CopyRecord record: copyRecords){
			result.add(record.getCopy());
		}
		for (CopyRecord record: splitRecords){
			result.add(record.getCopy());
		}
		return result;
	}
	
	
	
	
	
	
	
	
	private class CopyRecord{
		
		// the copy in which this record is representing
		private Copy copy = null;
		// in case of a split record, the reference copy is the counter/first part of the splitting. Note: not a 2 ways reference. 
		private Copy reference = null; 
		// the xpath keys that are the sources of the assignment, multiple due to possibility of concat. Constant assignment does not have 
		private List<String> fromKeys = new ArrayList<String>();
		// the xpath key that is the target of the assignment
		private String toKey = null;
		// whether or not this record is collided and swapped, 'swapped' means the to variable in the copy has been replaced by the temp variable for maintaining the orginal value.
		private boolean swapped = false;
		
		private CopyRecord(Copy aCopy){
			super();
			setCopy(aCopy);
			initializeKeys();
		}
		
		private void initializeKeys(){
			EObject obj = mapRule.getObjectMappedToCopy(copy);
			if (obj == null) return; // no keys are needed for split record
			if (obj instanceof StructuredOpaqueExpression){
				Expression exp = ((StructuredOpaqueExpression)obj).getExpression();
				if (exp instanceof ModelPathExpression){
					ModelPathExpression mpe = (ModelPathExpression)exp;
					extractFromKeys(mpe);
					extractToKey(mpe);
				}
				else if (exp instanceof FunctionExpression){
					FunctionExpression fx = (FunctionExpression)exp;
					extractFromKeysForFnExp(fx);
					extractToKeyForFnExp(fx);
				}
			}
			else if (obj instanceof LiteralSpecification){
				// there is no from key needed for constant assignment
				extractToKey((LiteralSpecification)obj);
			}
		}
		
		private void extractToKeyForFnExp(FunctionExpression fx){
			String funcId = fx.getDefinition().getFunctionID();
			if (CoreFunctions.CONCAT_FUNCTION.equals(funcId)){
				extractToKey(fx);
				return;
			}
			else if (CoreFunctions.SET_INDEX_OF_COLLECTION_FUNCTION.equals(funcId)){
				StringBuffer sb = new StringBuffer();
				sb.append(copy.getTo().getVariable().getName());				
				FunctionArgument outputFnArg = (FunctionArgument)fx.getArguments().get(0);
				ModelPathExpression toMpe = (ModelPathExpression)outputFnArg.getArgumentValue();
				String path = ConcatFunctionSerializer.getQueryString(toMpe, mapRule);
				if (path != null && ! path.equals("")){
					sb.append(XPathExpressionSerializer.STEP_SEP);
					sb.append(path);			
				}
				toKey = sb.toString();
			}
		}
		
		private void extractFromKeysForFnExp(FunctionExpression fx){
			String funcId = fx.getDefinition().getFunctionID();
			if (CoreFunctions.CONCAT_FUNCTION.equals(funcId)){
				for (Iterator<?> i = fx.getArguments().iterator(); i.hasNext();){
					FunctionArgument fArg = (FunctionArgument)i.next();
					Expression argExp = fArg.getArgumentValue();
					if (argExp instanceof FunctionExpression){
						extractFromKeysForFnExp((FunctionExpression)argExp);
					}
					else if (argExp instanceof ModelPathExpression){
						extractFromKeys((ModelPathExpression)argExp);
					}
					// we'll ignore StringLiteralExpression as an fromKey
				}
			}
			else if (CoreFunctions.SET_INDEX_OF_COLLECTION_FUNCTION.equals(funcId)){
				FunctionArgument inputFnArg = (FunctionArgument)fx.getArguments().get(2);
				Expression argExp = inputFnArg.getArgumentValue();
				if (argExp instanceof FunctionExpression){
					extractFromKeysForFnExp((FunctionExpression)argExp);
				}
				else if (argExp instanceof ModelPathExpression){
					extractFromKeys((ModelPathExpression)argExp);
				}
			}
		}
		
		// for MPE, LiteralSpec, concat function
		private void extractToKey(EObject valueSpec){
			ModelPathExpression toMpe = mapRule.createToMPE(valueSpec);
			StringBuffer sb = new StringBuffer();
			if (copy.getTo().getVariable() != null)
				sb.append(copy.getTo().getVariable().getName());
			String path = ConcatFunctionSerializer.getQueryString(toMpe, mapRule);
			if (path != null && ! path.equals("")){
				sb.append(XPathExpressionSerializer.STEP_SEP);
				sb.append(path);
			}
			toKey = sb.toString();
		}
		
		private void extractFromKeys(ModelPathExpression mpe){
			StringBuffer sb = new StringBuffer();
			if (copy.getFrom().getVariable() != null){
				sb.append(copy.getFrom().getVariable().getName());
			}
			else{
				String varName = null;
				if (mpe.getSteps().get(0) instanceof ReferenceStep){
					EObject refObj = ((ReferenceStep)mpe.getSteps().get(0)).getReferencedObject();
					if (refObj instanceof ObjectPin){
						BPELVariable var = pin2var.get(refObj);
						if (var !=  null){
							varName = var.getName();
						}
					}
				}
				if (varName == null)
					return;	// No name found, skip it.
				else
					sb.append(varName);
			}
			String path = ConcatFunctionSerializer.getQueryString(mpe, mapRule);
			if (path != null && ! path.equals("")){
				sb.append(XPathExpressionSerializer.STEP_SEP);
				sb.append(path);
			}
			fromKeys.add(sb.toString());
		}
		
		private boolean isSwapped(){
			return this.swapped;
		}
		
		private void setSwapped(boolean b){
			swapped = b;
		}
		
		private List<String> getFromKeys(){
			return fromKeys;
		}
		
		private String getToKey(){
			return toKey;
		}
		
		private Copy getCopy() {
			return copy;
		}

		private void setCopy(Copy copy) {
			this.copy = copy;
		}

		private Copy getReference() {
			return reference;
		}

		private void setReference(Copy reference) {
			this.reference = reference;
		}

		private void swapTarget(){
			BPELVariable var = (BPELVariable)copy.getTo().getVariable();
			BPELVariable tempVar = mapRule.getTempVariable(var);
			copy.getTo().setVariable(tempVar);
			setSwapped(true);
		}
		
		private boolean isPassThrough(){
			return getFromKeys().size() == 1 && getFromKeys().contains(getToKey());
		}
	}
	
}


