package com.ibm.btools.te.ilm.heuristics.helper;
/*
* Licensed Materials - Property of IBM
* 5724-I74, 5724-I75
*  Copyright IBM Corporation 2003, 2009. All rights reserved.
* US Government Users Restricted Rights - Use, duplication or disclosure
* restricted by GSA ADP Schedule Contract with IBM Corp.
*/
import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;

import com.ibm.btools.bom.model.artifacts.NamedElement;
import com.ibm.btools.bom.model.processes.activities.ActivityNode;
import com.ibm.btools.bom.model.processes.activities.ConnectableNode;
import com.ibm.btools.bom.model.processes.activities.InitialNode;
import com.ibm.btools.bom.model.processes.activities.InputControlPin;
import com.ibm.btools.bom.model.processes.activities.InputObjectPin;
import com.ibm.btools.bom.model.processes.activities.InputPinSet;
import com.ibm.btools.bom.model.processes.activities.LoopNode;
import com.ibm.btools.bom.model.processes.activities.OutputControlPin;
import com.ibm.btools.bom.model.processes.activities.OutputObjectPin;
import com.ibm.btools.bom.model.processes.activities.OutputPinSet;
import com.ibm.btools.bom.model.processes.activities.Pin;
import com.ibm.btools.bom.model.processes.activities.StructuredActivityNode;
import com.ibm.btools.bom.model.processes.activities.util.ActivitiesSwitch;

/**
 *
 * This class walks through a BOM model, stopping at the next object of a particular type.
 * 
 */
public class BomWalker {

	public static final String copyright = "Licensed Material - Property of IBM " + " 5724-I74, 5724-I75 " + "(C) Copyright IBM Corporation 2003, 2009. All Rights Reserved. " + "U.S. Government Users Restricted Rights - Use, duplication or disclosure " + "restricted by GSA ADP Schedule Contract with IBM Corp.".intern(); //$NON-NLS-1$ //$NON-NLS-2$ //$NON-NLS-3$ //$NON-NLS-4$ //$NON-NLS-5$
	
	/* These are the constants that are used in the parameter to the next(...) method to 
	 * indicate what the destination BOM object type is
	 */
	public static final int INPUT_SET = 0;
	public static final int OUTPUT_SET = 1;
	public static final int OUTPUT_SET_FROM_PROCESS_INPUTSET = 10;

	/* This is used to delegate to the proper method, given the type of incoming object */
	private BOMWalkerSwitch bomSwitch = null;
	
	/* These attributes are used to pass information into the BOMWalkerSwitch, since the 
	 * API doesn't support it directly.	 */
	private int fDestType;
	private boolean fDoJump;

	/**
	 * This class uses the EMF ActvitiesSwitch to perform processing particular to the each 
	 * relevant type of EObject.  Note, this is necessary because of EMF's allowance for multiple
	 * inheritance and the way Interfaces are used.
	 */
	private class BOMWalkerSwitch extends ActivitiesSwitch {

		public Object caseOutputPinSet(OutputPinSet object) {
			return getNextInternal(object, fDestType, fDoJump);
		}
		
		public Object caseInputPinSet(InputPinSet object) {
			return getNextInternal(object, fDestType, fDoJump);
		}

	}

	/**
	 * Default constructor.
	 */
	public BomWalker(){
		bomSwitch = new BOMWalkerSwitch();
	}
	
	/**
	 * This method is the same as calling <code>next(startingPlace, destType, true)</code>
	 */
	public List getNext(NamedElement startingPlace, int destType){
		return getNext(startingPlace, destType, true);
	}
	
	/**
	 * This method returns the next element in the BOM process graph that is of type described
	 * by <code>destType</code>.  If <code>doJump</code> is true, then the walker will
	 * jump over any exclusive or concurrent branches if the starting point is a branch point, or
	 * the internals of a SAN if he starting place is a SAN input set.
	 * 
	 * @param startingPlace The BOM object from which to start walking
	 * @param destType The type of BOM object to stop at.  Must be either 
	 * <code>INPUT_SET</code> or <code>OUTPUT_SET</code>
	 * @param doJump When true, all branches and SAN internals are jumped over if the starting place
	 * is a branch point.
	 * @return List - The list of next BOM objects of the type described by <code>destType</code>,
	 * or <code>null</code> if no such element can be reached from the <code>startingPlace</code>
	 */
	public List getNext(NamedElement startingPlace, int destType, boolean doJump){
		fDestType = destType;
		fDoJump = doJump;
		return (List)bomSwitch.doSwitch(startingPlace);

	}

	/**
	 * Starting from an output set, this method returns the object of type destType.  It does not
	 * cross end-line SAN boundaries, and will return an empty list or null if it reaches that case.
	 * 
	 * @param outputSetStart The output set to start walking from
	 * @param destType The type of object to stop at (either INPUT_SET or OUTPUT_SET)
	 * @param doJump Whether or not to jump over intermediate branch regions / nested SANs
	 * @return List - The list of objects of type <code>destType</code> that are encountered.
	 */
	private List getNextInternal(OutputPinSet outputSetStart, int destType, boolean doJump){

		List results = null;

		switch (destType){
			
			case INPUT_SET:
				if (doJump){
					/* Jumping */
					if (BomHelper.getInstance().isCBranch(outputSetStart)){
						/* We need to jump over the entire CBR */
						NamedElement mp = BomHelper.getInstance().findConcurrentMergePoint(outputSetStart);
						if (mp instanceof InputPinSet){
							results = new ArrayList();
							results.add(mp);
						} else {
							/* The merge point is the 'back wall' of the containing SAN, so results 
							 * is an empty list since we would have to cross SAN boundaries to get to 
							 * the next input set
							 */
							//results = new ArrayList();
							results = null;
						}
					} else {
						/* There is nothing to jump since this is not a branch point, so treat 
						 * this as the normal, non-jumping case. */
						results = getNextInternal(outputSetStart, INPUT_SET, false);
					}

				} else {
					/* No jumping */
					results = BomUtils.getNextInputPinSets(outputSetStart);
				}
				
				break;
				
			case OUTPUT_SET:
			
				if (doJump){
					/* Jumping */
					
					/* Q: If the starting output set is a CBP, and then the CMP (input set) is actually
					 * an EBP, do we return the output set on the CMP or the EMP? 
					 * A: EMP, since our logic
					 * is to jump at every possible step.  In order to cross from an input set to an output
					 * set, is possible that we could be crossing a EBR boundary. 
					 */
					if (BomHelper.getInstance().isCBranch(outputSetStart)){
						NamedElement cmp = BomHelper.getInstance().findConcurrentMergePoint(outputSetStart);
						if (cmp instanceof InputPinSet){
							// Q: Do we really want to continue to jump EBs / CBs indefinitely?
							// A: Yes, since that's the algorithm.  It won't happen if you choose to stop at
							//    an input set instead of an output set, so that's always an option.
							results = getNextInternal((InputPinSet)cmp, OUTPUT_SET, true);
						} else if (cmp instanceof StructuredActivityNode){
							/* The mp is the back-wall of the SAN, so return the empty set, since we don't 
							 * cross SAN boundaries.  (The SAN output sets are considered in a different region)
							 */
							results = new ArrayList();
						}

					} else {
						/* Not a CBP, so follow the output set to the next input sets and get the output sets
						 * corresponding to each of them */
						results = new ArrayList();
						List inputSets = BomUtils.getNextInputPinSets(outputSetStart);
						List nextInternals;
						for (Iterator it=inputSets.iterator(); it.hasNext(); ){
							nextInternals = getNextInternal((InputPinSet)it.next(), OUTPUT_SET, true);
							if (nextInternals!=null && !nextInternals.isEmpty()) {
								results.addAll(nextInternals);							
							}
						}
					}

				} else {
					/* No jumping */
					List inputSets = BomUtils.getNextInputPinSets(outputSetStart);
				
					if (inputSets.isEmpty()){
						/* We have reached the back wall of the SAN, so return the empty set */
						results = new ArrayList();
						
					} else {
						/* For each input set, walk to the first output set without jumping. */
						List outputSets = null;
						results = new ArrayList();
						for (Iterator it = inputSets.iterator(); it.hasNext(); ){
							outputSets = getNextInternal((InputPinSet)it.next(), OUTPUT_SET, false);

							if (outputSets!=null && !outputSets.isEmpty()) {
								/* First remove the ones we already have so we don't end up with duplicates */
								results.removeAll(outputSets);
								results.addAll(outputSets);
							}
						}
					}
				}
				
				break;
		}
		
		return results;
	}


	/**
	 * Starting from an input set, this method returns the next object of type destType.  It does not
	 * cross end-line SAN boundaries, and will return an empty list if it reaches that case.
	 * 
	 * @param inputSetStart The InputPinSet to start walking from
	 * @param destType The type of object to stop at (either INPUT_SET or OUTPUT_SET)
	 * @param doJump Whether or not to jump over intermediate branch regions or nested SANs
	 * @return List - The list of objects of type <code>destType</code> that were first encountered 
	 */
	private List getNextInternal(InputPinSet inputSetStart, int destType, boolean doJump){

		List results = null;

		switch (destType){
			
			case INPUT_SET:
				if (doJump){
					/* With jumping */
					if (BomHelper.getInstance().isEBranch(inputSetStart)){
						/* We want to return the first input set that follows this EMP, and we 
						 * want to jump in this case since the EMP could be a CBP in which case the 
						 * CBR would be contained within the EBR
						 * 
						 * Note: If the EMP is a containing SAN (meaning that the actual merge point was
						 * the 'back wall' of the SAN, we don't want to go any further, since we don't 
						 * cross SAN boundaries.
						 */

						NamedElement mp = BomHelper.getInstance().findExclusiveMergePoint(inputSetStart);
						if (mp instanceof OutputPinSet){
							/* mp is not the 'back wall' of the containing SAN. Proceed normally. */							 
							results = getNextInternal((OutputPinSet)mp, INPUT_SET, true);

						} else if (mp instanceof StructuredActivityNode){
							/* mp IS the 'back wall' of the containing SAN, so there are no more
							 * input sets after inputSetStart unless we cross SAN boundaries, so 
							 * return an empty list */
							results = new ArrayList();
						}
						
					} else {
						/* We now know that inputSetStart is not a EBP, but it could still be a CBP */

						/* Given that it's not an EBP, there must be only a single output set correlated
						 * to inputSetStart.
						 */
						if (!inputSetStart.getOutputPinSet().isEmpty()){
							OutputPinSet outputSet = (OutputPinSet)inputSetStart.getOutputPinSet().get(0);
							results = getNextInternal(outputSet, INPUT_SET, true);							
						
						} else {
							/* There was no correlated output set, so no next INPUT_SETs! */
							//results = new ArrayList();
							List opss = BomUtils.getCorrelatedPinSets(inputSetStart);
							if (opss.size()>0) {
								OutputPinSet outputSet = (OutputPinSet)opss.get(0);
								results = getNextInternal(outputSet, INPUT_SET, true);							
							}
						}
					}
					
				} else {
					/* Not jumping */
					results = new ArrayList();
					
					if (BomUtils.isProcess(inputSetStart.getAction())
							|| (inputSetStart.getAction() instanceof LoopNode)){
						/* This is a nested process or a loop node, so we want to step INSIDE */
						results.addAll(getNextInputPinSetsInSAN(inputSetStart));				
								
					} else {
						/* We are not on a nested process, just move to the next input sets by first
						 * finding the correlated output sets, then following the edges
						 */
						List tempInputPinSets = null;
						//for (Iterator it = inputSetStart.getOutputPinSet().iterator(); it.hasNext(); ){
						for (Iterator it = BomUtils.getCorrelatedPinSets(inputSetStart).iterator(); it.hasNext(); ){
							tempInputPinSets = BomUtils.getNextInputPinSets((OutputPinSet)it.next());
							
							/* First remove the ones we already have so we don't end up with duplicates */
							results.removeAll(tempInputPinSets);
							results.addAll(tempInputPinSets);
						}
					}					
				}

				break;
				
			case OUTPUT_SET:
				if (doJump){
					/* Jumping */
					OutputPinSet possibleCBP = null;
					
					if (BomHelper.getInstance().isEBranch(inputSetStart)){
						/* We want to return the output set that follows this EBP, and we 
						 * want to jump in this case since the EMP could be a CBP in which case the 
						 * CBR would be contained within the EBR
						 * 
						 * Note: If the EMP is a containing SAN (meaning that the actual merge point was
						 * the 'back wall' of the SAN, we don't want to go any further, since we don't 
						 * cross SAN boundaries.
						 */
						NamedElement mp = BomHelper.getInstance().findExclusiveMergePoint(inputSetStart);
						if (mp instanceof OutputPinSet){
							/* mp is not the 'back wall' of the containing SAN, but it still could be a CBP,
							 * set possibleCBP so that we know to check for this case later... */
							possibleCBP = (OutputPinSet)mp;

						} else if (mp instanceof StructuredActivityNode){
							/* The mp is the back-wall of the SAN, so return an empty list */
							results = new ArrayList();
						}
						
					} else {
						/* We now know that inputSetStart is not a EBP, but it could still be a CBP.  
						 * 
						 * NOTE: That doesn't matter because given that it's not an EBP, the next OUTPUT_SET 
						 * (which we will return) is the CBP.  If you jump from that output set, you would jump 
						 * over the CBR, but not if you just jump from this input set to that output set.
						 *  
						 * Given that it's not an EBP, there must be only a single output set correlated
						 * to inputSetStart.
						 */
						if (!inputSetStart.getOutputPinSet().isEmpty()){
							possibleCBP = (OutputPinSet)inputSetStart.getOutputPinSet().get(0);

//	TRIAL START
							 results = new ArrayList();
							 results.add(possibleCBP);
							 return results;							
//  TRIAL END							

							
						} else {
							/* There were no output sets correlated to the input set */
							//results = new ArrayList();
							List opss=BomUtils.getCorrelatedPinSets(inputSetStart);
							if (opss.size()>0) {
								possibleCBP = (OutputPinSet)opss.get(0); // Get the default outputPinSet
								results = new ArrayList();
								results.add(possibleCBP);
							}
							return results;														
						}
					}

					if (possibleCBP != null){
						/* We have to check if we are at a CBP.  If we are, we want to return 
						 * the output sets on the corresponding CMP. */
						if (BomHelper.getInstance().isCBranch((OutputPinSet)possibleCBP)){
							NamedElement cmp = BomHelper.getInstance().findConcurrentMergePoint((OutputPinSet)possibleCBP);
							if (cmp instanceof InputPinSet){
								// Q: Do we really want to continue to jump EBs / CBs indefinitely?
								// A: Yes, since that's the algorithm.  This would be avoided if we
								//    wanted to walk to the next INPUT_SET, instead of OUTPUT_SET
								results = getNextInternal((InputPinSet)cmp, OUTPUT_SET, true);
							} else if (cmp instanceof StructuredActivityNode){
								/* The mp is the back-wall of the SAN, so return an empty list */
								results = new ArrayList();
							}
	
						} else {
							/* Not a CBP, so the output set that we have reached is where we stop */
							results = new ArrayList();
							results.add(possibleCBP);
						}
					}
		
				} else {
					/* No jumping */
					
					/* If this is the input set of a SAN (nested process or loop node), step into 
					 * the SAN and return the first output set that we come to.	 */
					if (BomUtils.isProcess(inputSetStart.getAction())
							|| (inputSetStart.getAction() instanceof LoopNode)){
						results = new ArrayList();
						List inputSets = getNextInputPinSetsInSAN(inputSetStart);
						List outputSets = null;
						for (Iterator it = inputSets.iterator(); it.hasNext(); ){
							outputSets = getNextInternal((InputPinSet)it.next(), OUTPUT_SET, false);
							if (outputSets!=null && !outputSets.isEmpty()) {
								results.removeAll(outputSets);
								results.addAll(outputSets);
							}
						}
						
					} else {
						/* Just return the output sets that are correlated with the input set */
						results = new ArrayList();
						results.addAll(inputSetStart.getOutputPinSet());
					}
				}
				break;
			case OUTPUT_SET_FROM_PROCESS_INPUTSET:
				if (doJump){
					/* Jumping */
					
					/* Q: If the starting output set is a CBP, and then the CMP (input set) is actually
					 * an EBP, do we return the output set on the CMP or the EMP? 
					 * A: EMP, since our logic
					 * is to jump at every possible step.  In order to cross from an input set to an output
					 * set, is possible that we could be crossing a EBR boundary. 
					 */
//					if (BomHelper.getInstance().isCBranch(outputSetStart)){
//						NamedElement cmp = BomHelper.getInstance().findConcurrentMergePoint(outputSetStart);
//						if (cmp instanceof InputPinSet){
//							// Q: Do we really want to continue to jump EBs / CBs indefinitely?
//							// A: Yes, since that's the algorithm.  It won't happen if you choose to stop at
//							//    an input set instead of an output set, so that's always an option.
//							results = getNextInternal((InputPinSet)cmp, OUTPUT_SET, true);
//						} else if (cmp instanceof StructuredActivityNode){
//							/* The mp is the back-wall of the SAN, so return the empty set, since we don't 
//							 * cross SAN boundaries.  (The SAN output sets are considered in a different region)
//							 */
//							results = new ArrayList();
//						}
//
//					} else {
						/* Not a CBP, so follow the output set to the next input sets and get the output sets
						 * corresponding to each of them */
						results = new ArrayList();
						//List inputSets = BomUtils.getNextInputPinSets(outputSetStart);

						List inputSets = getNextInputPinSetsInSAN(inputSetStart);
						List nextInternals;
						for (Iterator it=inputSets.iterator(); it.hasNext(); ){
							nextInternals = getNextInternal((InputPinSet)it.next(), OUTPUT_SET, true);
							if (nextInternals!=null && !nextInternals.isEmpty()) {
								results.addAll(nextInternals);							
							}
						}
//					}
//
//				} else {
//					/* No jumping */
//					List inputSets = BomUtils.getNextInputPinSets(outputSetStart);
//				
//					if (inputSets.isEmpty()){
//						/* We have reached the back wall of the SAN, so return the empty set */
//						results = new ArrayList();
//						
//					} else {
//						/* For each input set, walk to the first output set without jumping. */
//						List outputSets = null;
//						results = new ArrayList();
//						for (Iterator it = inputSets.iterator(); it.hasNext(); ){
//							outputSets = getNextInternal((InputPinSet)it.next(), OUTPUT_SET, false);
//
//							if (outputSets!=null && !outputSets.isEmpty()) {
//								/* First remove the ones we already have so we don't end up with duplicates */
//								results.removeAll(outputSets);
//								results.addAll(outputSets);
//							}
//						}
//					}
				}

				break;
				
		}
		
		return results;
	}

	
	

	/**
	 * This is a utility method that gets the list of input sets that are reachable from the given 
	 * input set on an nested process.  It handles includes input sets that are attached to initial nodes
	 * in the containing SAN.
	 * 
	 * Note: This method has been updated to take into account the fact that initial nodes are 
	 * associated with one or input sets, and hence only the initial nodes that are associated with 
	 * the given sanInputPinSet are activated. 
	 * 
	 * @param sanInputPinSet The input set on the SAN from which we are walking
	 * @return List - The list of input sets that are contained in the SAN and are reachable from 
	 * <code>sanInputPinSet</code>
	 */
	private List getNextInputPinSetsInSAN(InputPinSet sanInputPinSet){
		List result = new ArrayList();
		List sanInputPins = new ArrayList(sanInputPinSet.getInputControlPin());
		sanInputPins.addAll(sanInputPinSet.getInputObjectPin());

		/* First add any input sets that are directly connected to pins in sanInputPinSet */
		for (int j=0; j<sanInputPins.size(); j++){
			if (((Pin)sanInputPins.get(j)).getOutgoing() != null){
				ConnectableNode cNode = ((Pin)sanInputPins.get(j)).getOutgoing().getTarget();
				if (cNode instanceof InputControlPin || cNode instanceof InputObjectPin){
					/* First remove all that already are in the list, so we don't get duplicates */
					result.removeAll(BomUtils.getInputPinSets((Pin)cNode));
					result.addAll(BomUtils.getInputPinSets((Pin)cNode));
				}
			}
		}
		
		/* Now add any input sets that are connected to by an initial node that is associated
		 * with sanInputPinSet (or that is not connected to any input sets) */
		List nodeContents = new ArrayList();
		if(sanInputPinSet.getAction() instanceof StructuredActivityNode){
			nodeContents = ((StructuredActivityNode)sanInputPinSet.getAction()).getNodeContents();
		}
		ActivityNode node = null;
		for (Iterator it = nodeContents.iterator(); it.hasNext(); ){
			node = (ActivityNode)it.next();
			if (node instanceof InitialNode && ( ((InitialNode)node).getEntryPoint().isEmpty() || ((InitialNode)node).getEntryPoint().contains(sanInputPinSet) )){
				if (((InitialNode)node).getOutgoing() != null){
					ConnectableNode cNode = ((InitialNode)node).getOutgoing().getTarget();
					if (cNode instanceof InputControlPin || cNode instanceof InputObjectPin){
						result.removeAll(BomUtils.getInputPinSets((Pin)cNode));
						result.addAll(BomUtils.getInputPinSets((Pin)cNode));
					}
				}
			}
		}
		return result;
	}		

	/**
	 * This is a utility method that gets the list of output sets on the containing SAN that are 
	 * reachable from the given output set that belongs to an action inside the SAN, 
	 * by following each pin from the output set.
	 * 
	 * @param outputSet The output set to start from.
	 * @return List The list of SAN output sets reachable from <code>outputSet</code>
	 */
	private List getNextOutputPinSets(OutputPinSet outputSet){
		List result = new ArrayList();
		List outputPins = new ArrayList(outputSet.getOutputObjectPin());
		List outputSets = null;
		outputPins.addAll(outputSet.getOutputControlPin());

		for (int j=0; j<outputPins.size(); j++){
			if (((Pin)outputPins.get(j)).getOutgoing() != null){
				ConnectableNode cNode = (ConnectableNode)((Pin)outputPins.get(j)).getOutgoing().getTarget();
				if (cNode instanceof OutputControlPin || cNode instanceof OutputObjectPin){
					outputSets = BomUtils.getOutputPinSets((Pin)cNode);
					/* First remove all that already are in the list, so we don't get duplicates */
					result.removeAll(outputSets);
					result.addAll(outputSets);
				}
			}
		}
		return result;
	}		
}
