package com.ibm.btools.te.ilm.heuristics.helper;
/*
* Licensed Materials - Property of IBM
* 5724-I74, 5724-I75
*  Copyright IBM Corporation 2003, 2009. All rights reserved.
* US Government Users Restricted Rights - Use, duplication or disclosure
* restricted by GSA ADP Schedule Contract with IBM Corp.
*/
import java.util.ArrayList;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;

import com.ibm.btools.bom.model.artifacts.NamedElement;
import com.ibm.btools.bom.model.processes.activities.InputPinSet;
import com.ibm.btools.bom.model.processes.activities.OutputPinSet;

/**
 * This class is used to keep track of the node paths that are involved in finding the merge
 * point for a given branch point.
 */
public class NodePathContext {

	public static final String copyright = "Licensed Material - Property of IBM " + " 5724-I74, 5724-I75 " + "(C) Copyright IBM Corporation 2003, 2009. All Rights Reserved. " + "U.S. Government Users Restricted Rights - Use, duplication or disclosure " + "restricted by GSA ADP Schedule Contract with IBM Corp.".intern(); //$NON-NLS-1$ //$NON-NLS-2$ //$NON-NLS-3$ //$NON-NLS-4$ //$NON-NLS-5$
	
	/* The number of node paths that are currently alive */
	private int fPathCount;
	
	/* The map of Node -> List of NodePath, where the number of NodePaths in the list is the
	 * number that have visited that node */
	private HashMap fNodeMap;
	
	/**
	 * Default constructor
	 */
	public NodePathContext(){
		fPathCount = 0;
		fNodeMap = new HashMap();
	}
	
	/**
	 * Registers the given NodePath in the list that must merge before a merge point is detected
	 * 
	 * @param nodePath The NodePath to register
	 */
	public void registerNodePath(NodePath nodePath){
		fPathCount++;
		
		/* Go through the node path and record each of the places it has been. */
		Object node;
		for (Iterator it = nodePath.getNodeHistory().iterator(); it.hasNext(); ){
			recordNodeToMap(it.next(), nodePath);
		}
	}

	/**
	 * Records each nodes visited since this node path was last recorded
	 * 
	 * @param nodePath The NodePath that contains the nodes to record
	 */
	public void updateNodePathRecord(NodePath nodePath){
		Iterator it = nodePath.getLastVisited().iterator();
		while (it.hasNext()){
			recordNodeToMap(it.next(), nodePath);
		}		
	}
	
	/**
	 * Record the give node in the map of places we've been
	 * @param node
	 * @param nodePath
	 */
	private void recordNodeToMap(Object node, NodePath nodePath){
//		if (node instanceof OutputPinSet){
//			/* This must be the output set at the end of the containing SAN, so record the SAN itself,
//			 * since other node paths could reach different output sets on the same SAN, but we want
//			 * to match these two types of cases together...
//			 */
//			node = ((OutputPinSet)node).getAction();			
//		}

		if (fNodeMap.containsKey(node)){
			List nodePathList = (List)fNodeMap.get(node);
			nodePathList.add(nodePath);
//			fNodeMap.put(node, new Integer(((Integer)fNodeMap.get(node)).intValue() + 1));
			fNodeMap.put(node, nodePathList);
		} else {
//			fNodeMap.put(node, new Integer(1));
			List nodePathList = new ArrayList();
			nodePathList.add(nodePath);
			fNodeMap.put(node, nodePathList);
		}
		
	}
	
	/**
	 * This method determines whether a merge point has been discovered, and returns it
	 * if it has.  If not, it returns null.
	 * @return NamedElement - The merge point, or null if a merge point has not yet been discovered
	 */
	public NamedElement findMerge(){

		int magicInt = fPathCount;
		Iterator it = fNodeMap.keySet().iterator();
		Object key = null;
		while (it.hasNext()){
			key = it.next();
			if ( ((List)fNodeMap.get(key)).size() == magicInt){
				return (NamedElement)key;
			}
		}
		
		/* Couldn't find a node that has been visited by all the paths */
		return null;
		
	}
	
	/**
	 * Once a merge point has been found, this method returns the branch region closed by the
	 * merge point.  The branch region is the list of elements that were visited en route to 
	 * finding (and including) the merge point, starting from the branch point.
	 * 
	 * @return List - The list of <code>NamedElement</code>s in the branch region, or null
	 * if the merge point has not yet been identified by this NodePathContext.
	 */
	public List getBranchRegion(NamedElement element){
		/* The merge point is the element in the fNodeMap that maps to a list with the same
		 * number of NodePath entries as there are node paths in existence.  Every NodePath in 
		 * the list must have the merge point as an element in its node history, and every node
		 * that comes before that element in the node history was visited on the way to the 
		 * merge point.
		 * 
		 * Therefore, the set of nodes that together form the branch region can be found by 
		 * taking the union of all the nodes that come before the merge point in the node 
		 * history list of each NodePath mapped to by the merge point in fNodeMap.
		 * 
		 * In other words... 
		 * the branch region is made up of the set of nodes that come before
		 * the merge point in the node history list of each NodePath in the list of NodePaths 
		 * mapped to (in fNodeMap) by the merge point.  
		 */

		Object mergePoint = findMerge();
		if (mergePoint == null){
		 	return null;
		}
		
		List result = new ArrayList();
		List nodePaths = (List)fNodeMap.get(mergePoint);
		List nodePathHistory = null;				 
		List tempSubList = null;
		for (Iterator it = nodePaths.iterator(); it.hasNext(); ){
			nodePathHistory = ((NodePath)it.next()).getNodeHistory();
			/* The merge point must be in the history list, or else it wouldn't be the merge point */
			int mergePointIndex = nodePathHistory.indexOf(mergePoint);

			/* Add the merge point itself to the list */
			tempSubList = nodePathHistory.subList(0, mergePointIndex + 1);

			/* First remove all so that we will not end up with duplicates in the list */
			result.removeAll(tempSubList);
			result.addAll(tempSubList);
		}
		
		/* Now go through the resultant list of nodes in the branch region and add the corresponding
		 * input sets / output sets that should be considered to be in the branch region by nature of
		 * the fact that they are correlated.
		 */
		NamedElement node = null;
		NamedElement tempInputPinSet = null;
		List possibleInputPinSets = new ArrayList();
		List definiteInputPinSets = new ArrayList();
		List tempInputPinSets = new ArrayList();
		BomWalker bomWalker = new BomWalker();
		
		for (Iterator it = result.iterator(); it.hasNext(); ){
			node = (NamedElement)it.next();
			
			if (node instanceof OutputPinSet){
				/* We need to determine whether the input sets on either side of this output set should
				 * be included in the list of nodes contained in this branch region. The test is as follows:
				 * If at least one output set in the branch region has a pin whose target is a pin in the
				 * input set, AND at least one output set in the branch region has a pin whose source is
				 * any pin in the input set, then the input set belongs in the branch region.
				 * 
				 * This complicated check is necessary because it is not sufficient to simply add all the 
				 * input sets that are correlated with an output set that is already in the branch region
				 * -- it is possible that an output set is correlated to multiple input sets, only some 
				 * of which belong in the branch region.
				 * 
				 * First, load the possibleInputPinSets list with all the input sets that are reachable
				 * from the current output set.
				 */
				tempInputPinSets = bomWalker.getNext(node, BomWalker.INPUT_SET);
				if (tempInputPinSets != null){
					/* First, remove all so that we won't have duplicates */
					possibleInputPinSets.removeAll(tempInputPinSets);
					possibleInputPinSets.addAll(tempInputPinSets);
				}				 
			}
		}

		for (Iterator it = result.iterator(); it.hasNext(); ){
			node = (NamedElement)it.next();
			
			if (node instanceof OutputPinSet){
				tempInputPinSets = ((OutputPinSet)node).getInputPinSet();
				for (Iterator it2 = tempInputPinSets.iterator(); it2.hasNext(); ){
					tempInputPinSet = (InputPinSet)it2.next();
					if (possibleInputPinSets.contains(tempInputPinSet) && !definiteInputPinSets.contains(tempInputPinSet)){
						definiteInputPinSets.add(tempInputPinSet);
					}
				}
			}
		}

		result.addAll(definiteInputPinSets);
		
		/* Now go through one more time to check for nested branch regions.
		 * If the input set or output set is a nested branch point, include all the objects in 
		 * that branch region to be part of this branch region as well.
		 */
		List nestedBranchRegions = new ArrayList();
		List tempBranchRegion;
		for (Iterator it = result.iterator(); it.hasNext(); ){
			node = (NamedElement)it.next();
			
			/* We want to ignore the branch point for which this method was called in the first place */
			if (node.equals(element)){
				continue;
			}
			
			if ((node instanceof OutputPinSet && BomHelper.getInstance().isCBranch((OutputPinSet)node))
			    || (node instanceof InputPinSet && BomHelper.getInstance().isEBranch((InputPinSet)node))){

				tempBranchRegion = BomHelper.getInstance().getBranchRegion(node);
				nestedBranchRegions.removeAll(tempBranchRegion);
				nestedBranchRegions.addAll(tempBranchRegion);
			}
							
		}

		result.removeAll(nestedBranchRegions);
		result.addAll(nestedBranchRegions);
		
		return result;
	}
}
