/**
 * <copyright>
 *
 *    Copyright IBM Corporation 2003, 2009. All rights reserved.
 *
 * </copyright>
 *
 * %W%
 * @version %I% %H%
 */
package com.ibm.btools.te.ilm.heuristics.util;
/*
* Licensed Materials - Property of IBM
* 5724-I74, 5724-I75
*  Copyright IBM Corporation 2008, 2009. All rights reserved.
* US Government Users Restricted Rights - Use, duplication or disclosure
* restricted by GSA ADP Schedule Contract with IBM Corp.
*/


import java.math.BigInteger;
import java.util.Calendar;
import java.util.Date;
import java.util.List;

import com.ibm.btools.bom.model.artifacts.Comment;
import com.ibm.btools.bom.model.time.OffsetDuration;
import com.ibm.btools.bom.model.time.OffsetWeekDay;
import com.ibm.btools.bom.model.time.RecurringTimeIntervals;
import com.ibm.btools.bom.model.time.TimeInterval;
import com.ibm.btools.te.framework.TransformationRoot;
import com.ibm.btools.te.framework.TransformationRule;
import com.ibm.btools.te.framework.util.TransformationUtil;
import com.ibm.btools.te.ilm.heuristics.TimeIntervalsRule;
import com.ibm.btools.util.logging.LogHelper;
import com.ibm.wbit.businesscalendar.CalFactory;
import com.ibm.wbit.businesscalendar.DateType;
import com.ibm.wbit.businesscalendar.FreqType;
import com.ibm.wbit.businesscalendar.Recur;
import com.ibm.wbit.businesscalendar.ValueType;
import com.ibm.wbit.businesscalendar.Vevent;
import com.ibm.wbit.businesscalendar.XTypeType;

/**
 * Provides a set of helper methods for creating and updating TimeIntervals related artifacts
 */
public class TimeIntervalsUtil {
	public static final String copyright = "Licensed Material - Property of IBM " + " 5724-I74, 5724-I75 " + "(C) Copyright IBM Corporation 2003, 2009. All Rights Reserved. " + "U.S. Government Users Restricted Rights - Use, duplication or disclosure " + "restricted by GSA ADP Schedule Contract with IBM Corp.".intern(); //$NON-NLS-1$ //$NON-NLS-2$ //$NON-NLS-3$ //$NON-NLS-4$ //$NON-NLS-5$

	/**
	 * Create Vevent for this interval.
	 * @param timeInterval TimeInterval
	 * @param rti RecurringTimeIntervals
	 * @return Vevent
	 */
	public static Vevent createVevent(TimeInterval timeInterval, RecurringTimeIntervals rti) {
		if (LogHelper.isTraceEnabled()) {
			LogHelper.traceEntry(null, null, "createVevent", 
					"timeInterval -->, " + timeInterval, "com.ibm.btools.te.ilm");
		}

		if (timeInterval == null || rti == null) {
			return null;
		}
		
		Vevent vEvent = CalFactory.eINSTANCE.createVevent();		
		
		String intervalName = null;
		List comments = timeInterval.getOwnedComment();		
		if (comments != null && !comments.isEmpty()) {
			Comment comment = (Comment) comments.get(0);
			intervalName = comment.getBody();		
		}
		if (intervalName != null) {
			vEvent.setSummary(intervalName);
		} else {
			vEvent.setSummary(rti.getName() + TimeConstants.DOUBLE_UNDERSCORE);
		}
		
		if (timeInterval.getOffset() == null || timeInterval.getOffset().isEmpty()) {
			if (rti.getAnchorPoint() != null && rti.getAnchorPoint().getPointInTime() != null) {
				DateType startTime = CalFactory.eINSTANCE.createDateType();
				startTime.setValue(rti.getAnchorPoint().getPointInTime());
				startTime.setValue1(ValueType.DATETIME_LITERAL);
				vEvent.setDtstart(startTime);
			}
			if (vEvent.getDtstart() != null && timeInterval.getDuration() != null) {
				vEvent.setDuration(timeInterval.getDuration());
			}
		} else if (timeInterval.getOffset() != null && timeInterval.getOffset().size() == 1) {
			if (timeInterval.getOffset().get(0) instanceof OffsetDuration) {	// 1 offset in TimeInterval
				if (rti.getAnchorPoint() != null && rti.getAnchorPoint().getPointInTime() != null) {
					DateType startTime = CalFactory.eINSTANCE.createDateType();
					startTime.setValue(calculateTimeFromOffsetDuration(rti, (OffsetDuration) timeInterval.getOffset().get(0)));
					startTime.setValue1(ValueType.DATETIME_LITERAL);
					vEvent.setDtstart(startTime);
				}				
				if (vEvent.getDtstart() != null && timeInterval.getDuration() != null) {
					vEvent.setDuration(timeInterval.getDuration());
				}
			} else if (timeInterval.getOffset().get(0) instanceof OffsetWeekDay) {
				if (rti.getAnchorPoint() != null && rti.getAnchorPoint().getPointInTime() != null) {
					DateType startTime = CalFactory.eINSTANCE.createDateType();
					startTime.setValue(calculateTimeFromWeekDayOffset(rti, (OffsetWeekDay) timeInterval.getOffset().get(0)));
					startTime.setValue1(ValueType.DATETIME_LITERAL);
					vEvent.setDtstart(startTime);
				}				
				if (vEvent.getDtstart() != null && timeInterval.getDuration() != null) {
					vEvent.setDuration(timeInterval.getDuration());
				}
			} else {
				// error
				if (LogHelper.isTraceEnabled()) {
					LogHelper.traceExit(null, null, "*** Error in createVevent", 
							"timeInterval -->, " + timeInterval, "com.ibm.btools.te.ilm");
				}
				return null;
			}	
		} else {
			// error
			if (LogHelper.isTraceEnabled()) {
				LogHelper.traceExit(null, null, "*** Error in createVevent", 
						"timeInterval -->, " + timeInterval, "com.ibm.btools.te.ilm");
			}
			return null;
		}
		
		// add recurrence
		Recur recurrence = createRecurrence(rti);
		vEvent.getRrule().add(recurrence);
		
		// set attributes
		vEvent.setXType(XTypeType.ON_TIME_LITERAL);
		vEvent.setDescription("");
		
		if (LogHelper.isTraceEnabled()) {
			LogHelper.traceExit(null, null, "createVevent", 
					"timeInterval -->, " + timeInterval, "com.ibm.btools.te.ilm");
		}
			
		return vEvent;
	}
		
		/**
		 * Retrieves the startTime from the RecurringTimeIntervals.
		 * @param pointInTime
		 * @param durationOffset
		 * @return String
		 */
		private static String getStartTime(String pointInTime, String durationOffset) {
			if (LogHelper.isTraceEnabled()) {
				LogHelper.traceEntry(null, null, "getStartTime", 
						"pointInTime -->, " + pointInTime, "com.ibm.btools.te.ilm");
			}

			Calendar pointInTimeCalendar = TimeStringConverter.timeStringToTime(pointInTime);	
			Calendar adjustedCalendar = TimeStringConverter.calculateEndTime(pointInTimeCalendar, durationOffset);
			String startTime = TimeStringConverter.timeToTimeString(adjustedCalendar);

			if (LogHelper.isTraceEnabled()) {
				LogHelper.traceExit(null, null, "getStartTime", 
						"pointInTime -->, " + pointInTime, "com.ibm.btools.te.ilm");
			}
			return startTime;
		}

		private static Recur createRecurrence(RecurringTimeIntervals rti) {
			if (rti == null || rti.getRecurrencePeriod() == null) {
				return null;
			}
			
			Recur recurrence = CalFactory.eINSTANCE.createRecur();
			
			Duration duration = new Duration(rti.getRecurrencePeriod().getDuration());
			
			// only a unit is allowed in duration for recurrencePeriod
			
			if (duration.getYears() != 0) {
				recurrence.setFreq(FreqType.YEARLY_LITERAL);
				recurrence.setInterval(BigInteger.valueOf(duration.getYears()));
			} else if (duration.getWeeks() != 0) {
				recurrence.setFreq(FreqType.WEEKLY_LITERAL);
				recurrence.setInterval(BigInteger.valueOf(duration.getWeeks()));
			} else if (duration.getDays() != 0) {
				recurrence.setFreq(FreqType.DAILY_LITERAL);
				recurrence.setInterval(BigInteger.valueOf(duration.getDays()));
			} else if (duration.getHours() != 0) {
				recurrence.setFreq(FreqType.HOURLY_LITERAL);
				recurrence.setInterval(BigInteger.valueOf(duration.getHours()));
			} else if (duration.getMinutes() != 0) {
				recurrence.setFreq(FreqType.MINUTELY_LITERAL);
				recurrence.setInterval(BigInteger.valueOf(duration.getMinutes()));
			} else if (duration.getSeconds() != 0) {
				recurrence.setFreq(FreqType.SECONDLY_LITERAL);
				recurrence.setInterval(BigInteger.valueOf(duration.getSeconds()));
			}

			String numOfRecurrences = rti.getNumberOfRecurrences();
			if (numOfRecurrences != null && !numOfRecurrences.equals(TimeConstants.UNLIMITED_NATURAL)) {
				recurrence.setCount(new BigInteger(numOfRecurrences));
			}

			return recurrence;
	}
		
	// Calculate start time based on the offset duration	
	private static String calculateTimeFromOffsetDuration(RecurringTimeIntervals rti, OffsetDuration offsetDuration) {
		if (rti == null || offsetDuration == null) {
			return null;
		}
		String durationOffset = offsetDuration.getDuration();	
		return getStartTime(rti.getAnchorPoint().getPointInTime(), durationOffset);
	}

	
	/**
	 * Retrieves the TimeIntervalsRule from the TransformationRoot.
	 * @param rti RecurringTimeIntervals
	 * @param root TransformationRoot
	 * @return TimeIntervalsRule
	 */
	public static TimeIntervalsRule findTimeIntervalsRule (RecurringTimeIntervals rti, TransformationRoot root) {
		if (rti == null) {
			return null;
		}
		
		TimeIntervalsRule tiRule = null;
		TransformationRule rule = TransformationUtil.getRuleForSource(root, TimeIntervalsRule.class, rti);
		if (rule instanceof TimeIntervalsRule) {
			tiRule = (TimeIntervalsRule)rule;
		}
		return tiRule;
	}
	
	// Calculate start time based on the day of week offset
	private static String calculateTimeFromWeekDayOffset(RecurringTimeIntervals rti, OffsetWeekDay offsetWeekDay) {
		if (rti == null || offsetWeekDay == null) {
			return null;
		}
		Calendar pointInTimeCalendar = TimeStringConverter.timeStringToTime(rti.getAnchorPoint().getPointInTime());	
		Calendar adjustedCalendar = calculateIntervalStartTime(pointInTimeCalendar, offsetWeekDay);
		String convertedTime = TimeStringConverter.timeToTimeString(adjustedCalendar);
		
		return convertedTime;
	}

	// Calculate start time for TimeIntervals based on the day of the week
	private static Calendar calculateIntervalStartTime(Calendar startTime, OffsetWeekDay offsetWeekDay) {
		if (startTime == null) {
			return null;
		}
		
		if (offsetWeekDay == null) {
			return startTime;
		}
		
		Date intervalStartTime = null;
		// adjust day of week
		// adjust by number of weeks
		OffsetWeekDay offset = offsetWeekDay;
		
		int ordinalNumber;
		if (offset.getOrdinalNumber() != null) {
			ordinalNumber = offset.getOrdinalNumber().intValue();
		} else {
			ordinalNumber = 1;
		}
		
		Calendar calendar = Calendar.getInstance(startTime.getTimeZone());
		calendar.setTime(startTime.getTime());
		calendar.set(Calendar.MILLISECOND, 0);
		// find day of week and adjust by number of weeks
		for (int i=0; i<ordinalNumber; i++) {
			calendar.setTime(com.ibm.btools.te.ilm.heuristics.util.TimeStringConverter.findNextDayofWeek(calendar.getTime(), startTime.getTimeZone(), offset.getDay().getValue()));
			if (i < ordinalNumber - 1) {
				calendar.add(Calendar.DATE, 1);
			}
		}
		
		// adjust time
		if (offset.getOffsetTime() != null) {
			intervalStartTime = TimeStringConverter.calculateEndTime(calendar, offset.getOffsetTime()).getTime();
		} else {
			intervalStartTime = calendar.getTime();
		}

		calendar.setTime(intervalStartTime);
		if (LogHelper.isTraceEnabled()) {
			LogHelper.traceExit(null, null, "calculateEndTime", 
					"startTime -->, " + startTime, "com.ibm.btools.te.ilm");
		}
		return calendar;
	}
}
