package com.ibm.btools.te.ilm.heuristics.util;
/*
* Licensed Materials - Property of IBM
* 5724-I74, 5724-I75
*  Copyright IBM Corporation 2003, 2009. All rights reserved.
* US Government Users Restricted Rights - Use, duplication or disclosure
* restricted by GSA ADP Schedule Contract with IBM Corp.
*/

import java.util.Calendar;
import java.util.Date;
import java.util.GregorianCalendar;
import java.util.TimeZone;

import com.ibm.btools.util.converters.TimeZoneConverter;
import com.ibm.btools.util.logging.LogHelper;

/**
 * TimeStringConverter contains methods to convert a time string to a time format and vice-versa
 * Based on the TimeStringConverted from btools.util plugin.
 */
public class TimeStringConverter {
	
	public static final String copyright = "Licensed Material - Property of IBM " + " 5724-I74, 5724-I75 " + "(C) Copyright IBM Corporation 2003, 2009. All Rights Reserved. " + "U.S. Government Users Restricted Rights - Use, duplication or disclosure " + "restricted by GSA ADP Schedule Contract with IBM Corp.".intern(); //$NON-NLS-1$ //$NON-NLS-2$ //$NON-NLS-3$ //$NON-NLS-4$ //$NON-NLS-5$
	
	public final static int YEAR_INDEX   = 0;
	public final static int MONTH_INDEX  = 1;

	// week is never used
	public final static int WEEK_INDEX   = -2;
	
	public final static int DAY_INDEX    = 2;
	public final static int HOUR_INDEX   = 3;
	public final static int MINUTE_INDEX = 4;
	public final static int SECOND_INDEX = 5;	
	
	public final static int MONDAY_INDEX    = 0;
	public final static int TUESDAY_INDEX   = 1;
	public final static int WEDNESDAY_INDEX = 2;
	public final static int THURSDAY_INDEX  = 3;
	public final static int FRIDAY_INDEX    = 4;
	public final static int SATURDAY_INDEX  = 5;
	public final static int SUNDAY_INDEX    = 6;
	
	public final static String PERIOD_KEY = "P";
	public final static String YEAR_KEY   = "Y";
	public final static String MONTH_KEY  = "M";
	public final static String WEEK_KEY   = "W";
	public final static String DAY_KEY    = "D";
	public final static String TIME_KEY   = "T";
	public final static String HOUR_KEY   = "H";
	public final static String MINUTE_KEY = "M";
	public final static String SECOND_KEY = "S";
	
	public final static String DATE_SEPARATOR = "-";
	public final static String TIME_SEPARATOR = ":";
	public final static String UTC_TIME_ZONE_INDICATOR = "Z";
	public final static String ZONE_OFFSET_POS_SIGN = "+";
	public final static String ZONE_OFFSET_NEG_SIGN = "-";

	/**
	 * Concert a time to time string 
	 * @param calendar	 
	 * @return String
	 */
	public static String timeToTimeString(Calendar calendar) {
		if (LogHelper.isTraceEnabled()) {
			LogHelper.traceEntry(null, null, "timeToTimeString", 
					"Calendar -->, " + calendar, "com.ibm.btools.te.ilm");
		}

		StringBuffer sb = new StringBuffer();

		sb.append(calendar.get(Calendar.YEAR));
		sb.append(DATE_SEPARATOR);
		sb.append(padWithZero(calendar.get(Calendar.MONTH) + 1));
		sb.append(DATE_SEPARATOR);
		sb.append(padWithZero(calendar.get(Calendar.DATE)));
		sb.append(TIME_KEY);
		 
		sb.append(padWithZero(calendar.get(Calendar.HOUR_OF_DAY)));
		
		sb.append(TIME_SEPARATOR);
		sb.append(padWithZero(calendar.get(Calendar.MINUTE)));
		sb.append(TIME_SEPARATOR);
		sb.append(padWithZero(calendar.get(Calendar.SECOND)));
		
		int timeZoneOffset = calendar.getTimeZone().getRawOffset();
		
		if (timeZoneOffset == 0) {
			
			sb.append(UTC_TIME_ZONE_INDICATOR);
		} else {
		
			if (timeZoneOffset > 0) {
				sb.append("+");
			} else if (timeZoneOffset < 0) {
				sb.append("-");
			}
		
			int hourOffset = Math.abs(timeZoneOffset / 3600000);
			int minuteOffset = Math.abs((timeZoneOffset % 3600000) / 60000);
		
			sb.append(padWithZero(hourOffset));
			sb.append(TIME_SEPARATOR);
			sb.append(padWithZero(minuteOffset));
		}
		if (LogHelper.isTraceEnabled()) {
			LogHelper.traceExit(null, null, "timeToTimeString", 
					"Calendar -->, " + calendar, "com.ibm.btools.te.ilm");
		}
		return sb.toString();
	}
	
	/**
	 * Pad numbers with 0 in front
	 * @param number
	 * @return padded number
	 */
	private static String padWithZero(int number) {
		if (number >= 0 && number < 10) {
			return "0" + number;
		} else {
			return new Integer(number).toString();
		}
	}
	
	/**
	 * Concert the time string to time 
	 * @param timeString	 
	 * @return Calendar
	 */	
	public static Calendar timeStringToTime(String timeString) {
		if (LogHelper.isTraceEnabled()) {
			LogHelper.traceEntry(null, null, "timeToTimeString", 
					"timeString -->, " + timeString, "com.ibm.btools.te.ilm");
		}
		Calendar calendar = Calendar.getInstance(TimeZone.getTimeZone(TimeZone.getAvailableIDs(0)[0]));
		
		int dateTimeStringEndIndex = Math.max(timeString.lastIndexOf(ZONE_OFFSET_POS_SIGN), timeString.lastIndexOf(ZONE_OFFSET_NEG_SIGN));
		dateTimeStringEndIndex = Math.max(dateTimeStringEndIndex, timeString.lastIndexOf(UTC_TIME_ZONE_INDICATOR));
		String dateTimeString = timeString.substring(0, dateTimeStringEndIndex);
		
		String yearString = dateTimeString.substring(0, dateTimeString.indexOf(DATE_SEPARATOR));
		calendar.set(Calendar.YEAR, new Integer(yearString).intValue());
		
		String monthString = dateTimeString.substring(dateTimeString.indexOf(DATE_SEPARATOR) + 1, dateTimeString.lastIndexOf(DATE_SEPARATOR));
		calendar.set(Calendar.MONTH, new Integer(monthString).intValue() - 1);
		
		String dateString = dateTimeString.substring(dateTimeString.lastIndexOf(DATE_SEPARATOR) + 1, dateTimeString.indexOf(TIME_KEY));
		calendar.set(Calendar.DATE, new Integer(dateString).intValue());
		
		String hourString = dateTimeString.substring(dateTimeString.indexOf(TIME_KEY) + 1, dateTimeString.indexOf(TIME_SEPARATOR));
		calendar.set(Calendar.HOUR_OF_DAY, new Integer(hourString).intValue());
		
		String minuteString = dateTimeString.substring(dateTimeString.indexOf(TIME_SEPARATOR) + 1, dateTimeString.lastIndexOf(TIME_SEPARATOR));
		calendar.set(Calendar.MINUTE, new Integer(minuteString).intValue());
		
		String secondString = dateTimeString.substring(dateTimeString.lastIndexOf(TIME_SEPARATOR) + 1);
		int decimalIndex = secondString.indexOf('.'); 
		if (decimalIndex == -1) {
			calendar.set(Calendar.SECOND, new Integer(secondString).intValue());
			calendar.set(Calendar.MILLISECOND, 0);
		} else {
			String milliSecondString = secondString.substring(decimalIndex + 1);
			secondString = secondString.substring(0, decimalIndex);
			calendar.set(Calendar.SECOND, new Integer(secondString).intValue());
			calendar.set(Calendar.MILLISECOND, new Integer(milliSecondString).intValue());
		} 	
		
		String zoneString = timeString.substring(dateTimeStringEndIndex);
		
		if (!zoneString.equals(UTC_TIME_ZONE_INDICATOR)) {
			String hourOffset = zoneString.substring(0, zoneString.indexOf(TIME_SEPARATOR));
			if (hourOffset.charAt(0) == '+') {
				hourOffset = hourOffset.substring(1);
			}
			String minuteOffset = zoneString.substring(zoneString.indexOf(TIME_SEPARATOR) + 1);
		
			int rawOffset = Math.abs(Integer.parseInt(hourOffset) * 3600000) + Math.abs(Integer.parseInt(minuteOffset) * 60000);
			if (hourOffset.indexOf('-') != -1) {
				rawOffset = rawOffset * -1;
			}
			calendar.setTimeZone(TimeZoneConverter.getTimeZone(rawOffset));				
		}
		if (LogHelper.isTraceEnabled()) {
			LogHelper.traceExit(null, null, "timeToTimeString", 
					"timeString -->, " + timeString, "com.ibm.btools.te.ilm");
		}
		return calendar;
	}

	/**
	 * Calculate the end time given a start time, a time unit and a quantity 
	 * @param startTime
	 * @param timeUnit
	 * @param quantity
	 * @return Calendar
	 */
	public static Calendar calculateEndTime(Date startTime, int timeUnit, int quantity) {
		if (LogHelper.isTraceEnabled()) {
			LogHelper.traceEntry(null, null, "calculateEndTime", 
					"startTime -->, " + startTime, "com.ibm.btools.te.ilm");
		}
		Calendar calendar = new GregorianCalendar();
		calendar.setTime(startTime);
		calendar.set(Calendar.MILLISECOND, 0);
		
		int fieldType = -1;
		if (timeUnit == YEAR_INDEX) {
			calendar.add(Calendar.YEAR, quantity);
		} else if (timeUnit == MONTH_INDEX) {
			calendar.add(Calendar.MONTH, quantity);
		} else if (timeUnit == WEEK_INDEX) {
			calendar.add(Calendar.DATE, quantity * 7);
		} else if (timeUnit == DAY_INDEX) {
			calendar.add(Calendar.DATE, quantity);
		} else if (timeUnit == HOUR_INDEX) {
			calendar.add(Calendar.HOUR_OF_DAY, quantity);
		} else if (timeUnit == MINUTE_INDEX) {
			calendar.add(Calendar.MINUTE, quantity);
		} else if (timeUnit == SECOND_INDEX) {
			calendar.add(Calendar.SECOND, quantity);
		}
		if (LogHelper.isTraceEnabled()) {
			LogHelper.traceExit(null, null, "calculateEndTime", 
					"startTime -->, " + startTime, "com.ibm.btools.te.ilm");
		}
		return calendar;
	}
	
	/**
	 * Calculate the end time given a start time and a duration time string 
	 * @param time
	 * @param timeString the time string that represents the duration
	 * @return Calendar
	 */
	public static Calendar calculateEndTime(Calendar startTime, String durationTimeString) {
		if (LogHelper.isTraceEnabled()) {
			LogHelper.traceEntry(null, null, "calculateEndTime", 
					"startTime -->, " + startTime, "com.ibm.btools.te.ilm");
		}
		Calendar calendar = Calendar.getInstance(startTime.getTimeZone());
		calendar.setTime(startTime.getTime());
		calendar.set(Calendar.MILLISECOND, 0);
		
		String newDurationTimeStr = null;
		if (durationTimeString != null) {		
			if (durationTimeString.startsWith(TimeConstants.MINUS)) {
				// the start time is BEFORE the beginning on
				newDurationTimeStr = durationTimeString.substring(1);
				
				Duration duration = new Duration(newDurationTimeStr);
				
				if (duration.getYears() > 0) {
					calendar.add(Calendar.YEAR, -1 * duration.getYears());
				}
				if (duration.getMonths() > 0) {
					calendar.add(Calendar.MONTH, -1 * duration.getMonths());
				}
				
				if (duration.getWeeks() > 0) {
					calendar.add(Calendar.DATE, -1 * duration.getWeeks() * 7);
				}
				
				if (duration.getDays() > 0) {
					calendar.add(Calendar.DATE, -1 * duration.getDays());
				}
				if (duration.getHours() > 0) {
					calendar.add(Calendar.HOUR_OF_DAY, -1 * duration.getHours());
				}
				if (duration.getMinutes() > 0) {
					calendar.add(Calendar.MINUTE, -1 * duration.getMinutes());
				}
				if (duration.getSeconds() > 0) {
					calendar.add(Calendar.SECOND, -1 * duration.getSeconds());
				}
	
			} else {		
				Duration duration = new Duration(durationTimeString);
						
				if (duration.getYears() > 0) {
					calendar.add(Calendar.YEAR, duration.getYears());
				}
				if (duration.getMonths() > 0) {
					calendar.add(Calendar.MONTH, duration.getMonths());
				}
				
				if (duration.getWeeks() > 0) {
					calendar.add(Calendar.DATE, duration.getWeeks() * 7);
				}
				
				if (duration.getDays() > 0) {
					calendar.add(Calendar.DATE, duration.getDays());
				}
				if (duration.getHours() > 0) {
					calendar.add(Calendar.HOUR_OF_DAY, duration.getHours());
				}
				if (duration.getMinutes() > 0) {
					calendar.add(Calendar.MINUTE, duration.getMinutes());
				}
				if (duration.getSeconds() > 0) {
					calendar.add(Calendar.SECOND, duration.getSeconds());
				}
			}
		}
		if (LogHelper.isTraceEnabled()) {
			LogHelper.traceExit(null, null, "calculateEndTime", 
					"startTime -->, " + startTime, "com.ibm.btools.te.ilm");
		}
		return calendar;
	}
	
	/**
	 * Returns the next day of week
	 * @param startTime
	 * @param timeZone 
	 * @param dayOfWeek
	 * @return Date
	 */
	public static Date findNextDayofWeek(Date startTime, TimeZone timeZone, int dayOfWeek) {
		if (LogHelper.isTraceEnabled()) {
			LogHelper.traceEntry(null, null, "findNextDayofWeek", 
					"startTime -->, " + startTime, "com.ibm.btools.te.ilm");
		}
		Calendar calendar = new GregorianCalendar(timeZone);
		calendar.setTime(startTime);
		calendar.set(Calendar.MILLISECOND, 0);
		
		// set start time to midnight
		calendar.set(Calendar.HOUR, 0);
		calendar.set(Calendar.MINUTE, 0);
		calendar.set(Calendar.SECOND, 0);
		calendar.set(Calendar.AM_PM, Calendar.AM);
		
		// Keep adding one day to time until day of week is found
		for (int i=0; i<7; i++) {
			
			// Java day of week and BOM day of week are different
			int calendarDayOfWeek = calendar.get(Calendar.DAY_OF_WEEK);
			if (calendarDayOfWeek == Calendar.SUNDAY) {
				calendarDayOfWeek = 6;
			} else {
				calendarDayOfWeek = calendarDayOfWeek - 2;
			}
			
			if (calendarDayOfWeek == dayOfWeek) {
				if (LogHelper.isTraceEnabled()) {
					LogHelper.traceExit(null, null, "findNextDayofWeek", 
							"startTime -->, " + startTime, "com.ibm.btools.te.ilm");
				}
				return calendar.getTime();
			}
			calendar.add(Calendar.DATE, 1);
		}
		if (LogHelper.isTraceEnabled()) {
			LogHelper.traceExit(null, null, "findNextDayofWeek", 
					"startTime -->, " + startTime, "com.ibm.btools.te.ilm");
		}
		return null;
	}
	
	/**
	 * Calculates the offset duration
	 * @param startTime
	 * @param anchorPoint	 
	 * @return Duration
	 */	
	public static Duration calculateOffsetDuration(Calendar startTime, Calendar anchorPoint){
		if (LogHelper.isTraceEnabled()) {
			LogHelper.traceEntry(null, null, "calculateOffsetDuration", 
					"startTime -->, " + startTime, "com.ibm.btools.te.ilm");
		}
		Duration duration = new Duration();
		long secDiff = (startTime.getTime().getTime() - anchorPoint.getTime().getTime())/1000;
		if ( secDiff < 0 ) {
			duration.setNegative( true );
		} 
		secDiff = Math.abs( secDiff );
		int daysDiff = Math.round( secDiff/86400);
		int hoursDiff = Math.round( (secDiff - daysDiff*86400)/3600 );
		int minsDiff = Math.round( (secDiff - daysDiff*86400 - hoursDiff*3600)/60 );
		secDiff = secDiff - daysDiff*86400 - hoursDiff*3600 - minsDiff*60;
		int secsDiff = (new Long( secDiff )).intValue();
		if ( startTime.getTimeZone().inDaylightTime( startTime.getTime() ) && !anchorPoint.getTimeZone().inDaylightTime( anchorPoint.getTime() ) ) {
			hoursDiff += 1;
		} else if ( !startTime.getTimeZone().inDaylightTime( startTime.getTime() ) && anchorPoint.getTimeZone().inDaylightTime( anchorPoint.getTime() ) ) {
			hoursDiff -= 1;
		}			

		duration.setDays(daysDiff);
		duration.setHours(hoursDiff);
		duration.setMinutes(minsDiff);
		duration.setSeconds(secsDiff);
		if (LogHelper.isTraceEnabled()) {
			LogHelper.traceExit(null, null, "calculateOffsetDuration", 
					"startTime -->, " + startTime, "com.ibm.btools.te.ilm");
		}
		return duration;		
	}

	/**
	 * Converts a given time (relative to a baseTime) to a time string representation
	 * @param time
	 * @param baseTime	 
	 * @return String
	 */	
	public static String convertTimetoTimeString(String time, String baseTime) {
		if (LogHelper.isTraceEnabled()) {
			LogHelper.traceEntry(null, null, "convertTimetoTimeString", 
					"time -->, " + time, "com.ibm.btools.te.ilm");
		}
		int indTimeSeperator = baseTime.indexOf(TIME_KEY);
		if (indTimeSeperator == -1) {
			//log errors or warnings
			if (LogHelper.isTraceEnabled()) {
				LogHelper.traceExit(null, null, "convertTimetoTimeString", 
						"time -->, " + time, "com.ibm.btools.te.ilm");
			}
			return null;
		}		
		String date = baseTime.substring(0, indTimeSeperator + 1);
		String result = date + time;
		if (LogHelper.isTraceEnabled()) {
			LogHelper.traceExit(null, null, "convertTimetoTimeString", 
					"time -->, " + time, "com.ibm.btools.te.ilm");
		}
		return result;
	}

	/**
	 * Retrieves a time from a time string representation
	 * @param timeString	 	 
	 * @return String
	 */	
	public static String getTimeFromTimeString(String timeString) {
		if (LogHelper.isTraceEnabled()) {
			LogHelper.traceEntry(null, null, "getTimeFromTimeString", 
					"timeString -->, " + timeString, "com.ibm.btools.te.ilm");
		}
		int indTimeSeperator = timeString.indexOf(TIME_KEY);
		if (indTimeSeperator == -1) {
			//log errors/warnings
			if (LogHelper.isTraceEnabled()) {
				LogHelper.traceExit(null, null, "getTimeFromTimeString", 
						"timeString -->, " + timeString, "com.ibm.btools.te.ilm");
			}
		}
		String time = timeString.substring(indTimeSeperator + 1,timeString.length());
		if (LogHelper.isTraceEnabled()) {
			LogHelper.traceExit(null, null, "getTimeFromTimeString", 
					"timeString -->, " + timeString, "com.ibm.btools.te.ilm");
		}
		return time;
	}	
}