package com.ibm.btools.te.ilm.rulehandler.mapping;
/*
* Licensed Materials - Property of IBM
* 5724-I74, 5724-I75
*  Copyright IBM Corporation 2003, 2009. All rights reserved.
* US Government Users Restricted Rights - Use, duplication or disclosure
* restricted by GSA ADP Schedule Contract with IBM Corp.
*/
import java.security.MessageDigest;
import java.security.NoSuchAlgorithmException;

import org.eclipse.emf.ecore.EObject;

import com.ibm.btools.bom.model.artifacts.Element;
import com.ibm.btools.bom.model.models.InformationModel;
import com.ibm.btools.bom.model.processes.activities.ObjectPin;
import com.ibm.btools.bom.model.processes.activities.RetrieveArtifactPin;
import com.ibm.btools.bom.model.processes.activities.StoreArtifactPin;
import com.ibm.btools.expression.model.FunctionArgument;
import com.ibm.btools.infrastructure.util.UIDGenerator;
import com.ibm.btools.model.resourcemanager.ResourceMGR;
import com.ibm.btools.te.ilm.LoggingUtil;
import com.ibm.btools.te.ilm.resource.MessageKeys;
import com.ibm.btools.te.ilm.resource.TEResourceBundleSingleton;
import com.ibm.btools.util.exception.BTRuntimeException;
import com.ibm.wsspi.sca.scdl.Component;
import com.ibm.wsspi.sca.scdl.DocumentRoot;
import com.ibm.wsspi.sca.scdl.Export;
import com.ibm.wsspi.sca.scdl.Import;
import com.ibm.wsspi.sca.scdl.InterfaceSet;
import com.ibm.wsspi.sca.scdl.wsdl.WSDLPortType;

/**
 * Utility class that provides methods for generating unique IDs.
 */
public class UIDUtil {
	public static final String copyright = "Licensed Material - Property of IBM " + " 5724-I74, 5724-I75 " + "(C) Copyright IBM Corporation 2003, 2009. All Rights Reserved. " + "U.S. Government Users Restricted Rights - Use, duplication or disclosure " + "restricted by GSA ADP Schedule Contract with IBM Corp.".intern(); //$NON-NLS-1$ //$NON-NLS-2$ //$NON-NLS-3$ //$NON-NLS-4$ //$NON-NLS-5$	
	//prefix for the ObjectDefinition UID.
	public static final String UID_OBJECT_PREFIX = "OBJ";
	public static final String DEFAULT_DATA_CATALOG_IDENTIFIER = "FID-00008";
	
	/**
	 * Returns the uid for this ObjectDefinition
	 * @param eObj
	 * @return
	 */
	public static String getObjectDefinitionUID(EObject eObj) {
		return getUID(UID_OBJECT_PREFIX,eObj);
	}
	/**
	 * Returns the ObjectDefinition UID for the ObjectDefinition based on the pair source/target objects.
	 * @param srcObj typically a BOM Object
	 * @param tarObj typically a BPEL/WSDL/XSD Object
	 * @return string uid
	 */
	public static String getObjectDefinitionUID(EObject srcObj, EObject tarObj) {
		return getUID(UID_OBJECT_PREFIX,srcObj,tarObj);		
	}

	/**
	 * Returns the <code>eObj</code> uid based on its EObject uid.
	 * @param eObj
	 * @return
	 */
	private static String getInternalUID(EObject eObj) {
		if (eObj instanceof Element) {
			String eObjUid = ((Element)eObj).getUid();
			return eObjUid;
		} else if (eObj instanceof FunctionArgument) {
			String eObjUid = ((FunctionArgument)eObj).getUid();
			return eObjUid;
		} else {
			return UIDGenerator.getUID("");
		}
	}
	/**
	 * Returns the unique <code>eObj</code> uid based using the given <code>prefix</code>
	 * @param prefix
	 * @param eObj
	 * @return
	 */
	public static String getUID(String prefix, EObject eObj) {
		if(eObj instanceof InformationModel ){
			//Algorithm for computing the default data catalog UID:
			//1. Default Data Cataog identifier: FID-00008
			//2.Project UID
			//3.Limitation: This will handle most user scenarios in that the user imports the entire project, where the project UID will retain. 
			//However, in the case of specific element like default catalog export and import the specific element into another project, we
			//will lose the traceability
			InformationModel im = (InformationModel)eObj;
			if(im.getUid().startsWith(DEFAULT_DATA_CATALOG_IDENTIFIER)){
				String projectName = ResourceMGR.getResourceManger().getProjectName(eObj);
				String hashKey = DEFAULT_DATA_CATALOG_IDENTIFIER;
				hashKey += MappingUtil.getProjectUID(projectName);
				return prefix+"-"+getMD5UID(hashKey);
			}
		}
		return prefix+"-"+getMD5UID(getInternalUID(eObj));
	}
	
	public static String getUID(String prefix, EObject srcObj, EObject tarObj) {
		String uid = getInternalUID(srcObj);
		//Synchronize the target OD UID with the default data catalog OD UID
		if(srcObj instanceof InformationModel ){			
			InformationModel im = (InformationModel)srcObj;
			if(im.getUid().startsWith(DEFAULT_DATA_CATALOG_IDENTIFIER)){
				String projectName = ResourceMGR.getResourceManger().getProjectName(srcObj);
				String hashKey = DEFAULT_DATA_CATALOG_IDENTIFIER;
				hashKey += MappingUtil.getProjectUID(projectName);
				uid = hashKey;
			}
		}
		return prefix+"-"+getMD5UID(uid+getUIDSuffix(tarObj));
	}
	
	/**
	 * Generate hex string with 32 length
	 * @param eObj
	 * @return
	 */
	public static String getUIDForBOMElement(EObject src, EObject target){	
		String uid = null;
		if(src instanceof ObjectPin && !(src instanceof RetrieveArtifactPin)&& !(src instanceof StoreArtifactPin)){
			ObjectPin pin = (ObjectPin)src;
			uid = getInternalUID(pin.getType());
		}else{
			uid = getInternalUID(src);
		}		
		return getMD5UID(uid+getUIDSuffix(target));
	}
	
	/**
	 * The UID Suffix is the instance class name of the target object (eObj) that particpate in a map.  
	 * This is based on the assumption that a src object only maps to a single object of one type. i.e.
	 * A src object won't map to more than one object with the same type.
	 * @param eObj
	 * @return
	 */
	private static String getUIDSuffix(EObject eObj) {
		String suffix = null;		
		if(eObj instanceof DocumentRoot){
			DocumentRoot docRoot = (DocumentRoot)eObj;
			if(docRoot.getComponent() != null){
				suffix = eObj.eClass().getInstanceClassName()+
				docRoot.getComponent().eClass().getInstanceClassName(); 
			}else if(docRoot.getExport() != null){
				suffix = eObj.eClass().getInstanceClassName()+
				docRoot.getExport().eClass().getInstanceClassName(); 
			}else if(docRoot.getImport() != null){
				suffix = eObj.eClass().getInstanceClassName()+
				docRoot.getImport().eClass().getInstanceClassName(); 
			}
		}else if(eObj instanceof WSDLPortType){
			if(eObj.eContainer() instanceof InterfaceSet){
				if(eObj.eContainer().eContainer() != null){
					suffix = eObj.eContainer().eContainer().eClass().getInstanceClassName()+
					eObj.eClass().getInstanceClassName();
				}
			}			
		}else{
			suffix =eObj.eClass().getInstanceClassName();
		}
		return suffix;
	}
	
	private static String getMD5UID(String key) {
        MessageDigest md5Algorithm = null;

        StringBuffer sbValueBeforeMD5 = new StringBuffer();

        try {
            md5Algorithm = MessageDigest.getInstance("MD5");
        } catch (NoSuchAlgorithmException e) {       	
			String params[]={"No MD5 Algorithm"};
			LoggingUtil.logError(MessageKeys.ERROR_GET_MD5_INSTANCE, params, e);
			BTRuntimeException runtimeEx = new BTRuntimeException(TEResourceBundleSingleton.INSTANCE.getMessage(MessageKeys.ERROR_GET_MD5_INSTANCE));
			runtimeEx.setChainedException(e);			
			throw runtimeEx;         
        }
        
        md5Algorithm.update(key.getBytes());

        byte[] array = md5Algorithm.digest();
        StringBuffer sb = new StringBuffer();
        for (int j = 0; j < array.length; ++j) {
        	int b = array[j] & 0xFF;
        	if (b < 0x10) sb.append('0');
            sb.append(Integer.toHexString(b));
        }

        return sb.toString();
	}
}
